#pragma once
#include <core/base_column.hpp>
#include <gpu/gpu_algorithms.hpp>
#include <utility>
#include <map>

namespace CoGaDB{

	class GPU_Column_Cache{
		public:
		typedef std::map<ColumnPtr,gpu::GPU_Base_ColumnPtr> Map;


		/* \brief if GPU column already exists, then return pointer to it, otherwise create new GPU column by transferring the data the GPU memory
			\detail if the GPU Buffer is full, then delete least recently used Column*/
		const gpu::GPU_Base_ColumnPtr getGPUColumn(ColumnPtr);

		static GPU_Column_Cache& instance();

		void printStatus() const throw();

		bool removeGPUColumn(gpu::GPU_Base_ColumnPtr);

		private:
		GPU_Column_Cache(unsigned int max_gpu_buffer_size_in_byte); //no constructor call outside of this class
		GPU_Column_Cache(const GPU_Column_Cache&); //no copy constructor
		GPU_Column_Cache& operator=(const GPU_Column_Cache&); //nocopy assignment
		Map map_;
		unsigned int max_gpu_buffer_size_in_byte_;
		bool caching_enabled_;
	};

}; //end namespace CoGaDB

