
#pragma once

#include <thrust/version.h>
#include <thrust/device_vector.h>
#include <thrust/iterator/zip_iterator.h>
#include <thrust/partition.h>
#include <thrust/copy.h>
#include <thrust/functional.h>
#include <thrust/sequence.h>
#include <thrust/generate.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>

#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>


using namespace std;

namespace CoGaDB{

	namespace gpu{
	

	
template <typename T>
struct Functor_Selection_Tuple_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  Functor_Selection_Tuple_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return ( thrust::get<0>(x)==a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Lesser : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Lesser(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)<a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Lesser_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Lesser_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)<=a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Greater : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Greater(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)>a);
  }
};

template <typename T>
struct Functor_Selection_Tuple_Greater_Equal : public thrust::unary_function<thrust::tuple<T,unsigned int>&,bool>
{
  const T a;
  
  Functor_Selection_Tuple_Greater_Equal(T _a) : a(_a) {}

  __host__ __device__
  bool operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return (thrust::get<0>(x)>=a);
  }
};

template <typename T>
struct Functor_Tuple_to_TID : public thrust::unary_function<thrust::tuple<T,unsigned int>&,unsigned int>
{
  __host__ __device__
  unsigned int operator()(const thrust::tuple<T,unsigned int>& x) const
  { 
    return thrust::get<1>(x); //return tid
  }
};

	}; //end namespace gpu

}; //end namespace CoGaDB

