#pragma once

#include <query_processing/definitions.hpp>
#include <util/getname.hpp>

namespace CoGaDB {
    namespace query_processing {
        namespace physical_operator {

            class CPU_Groupby_Operator : public hype::queryprocessing::UnaryOperator<TablePtr, TablePtr> {
            public:
                typedef hype::queryprocessing::OperatorMapper_Helper_Template<TablePtr>::TypedOperatorPtr TypedOperatorPtr;

                CPU_Groupby_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr child,
                        const std::list<std::string>& grouping_columns,
                        const std::list<ColumnAggregation>& aggregation_functions,
                        MaterializationStatus mat_stat = MATERIALIZE) : UnaryOperator<TablePtr, TablePtr>(sched_dec, child),
                grouping_columns_(grouping_columns),
                aggregation_functions_(aggregation_functions),
                mat_stat_(mat_stat) {
                }

                virtual bool execute() {
                    //std::cout << "Execute Groupby CPU" << std::endl;
                    //const TablePtr sort(TablePtr table, const std::string& column_name, SortOrder order=ASCENDING, MaterializationStatus mat_stat=MATERIALIZE, ComputeDevice comp_dev=CPU);
                    //this->result_=BaseTable::sort(this->getInputData(), column_name_,order_, mat_stat_,CPU);
                    this->result_ = BaseTable::groupby(this->getInputData(), grouping_columns_, aggregation_functions_, CPU);



                    //						std::cout << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << std::endl;
                    //						std::cout << "INPUT:" << std::endl;
                    //						this->getInputData()->print();
                    //						std::cout << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << std::endl;
                    //						std::cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
                    //						std::cout << "OUTPUT:" << std::endl;
                    //						this->result_->print();
                    //						std::cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
                    if (this->result_) {
                        setResultSize(((TablePtr) this->result_)->getNumberofRows());
                        return true;
                    } else
                        return false;
                }

                virtual ~CPU_Groupby_Operator() {
                }
            private:
                std::list<std::string> grouping_columns_;
                std::list<ColumnAggregation> aggregation_functions_;
                MaterializationStatus mat_stat_;
            };

            class GPU_Groupby_Operator : public hype::queryprocessing::UnaryOperator<TablePtr, TablePtr> {
            public:
                typedef hype::queryprocessing::OperatorMapper_Helper_Template<TablePtr>::TypedOperatorPtr TypedOperatorPtr;

                GPU_Groupby_Operator(const hype::SchedulingDecision& sched_dec,
                        TypedOperatorPtr child,
                        const std::list<std::string>& grouping_columns,
                        const std::list<ColumnAggregation>& aggregation_functions,
                        MaterializationStatus mat_stat = MATERIALIZE) : UnaryOperator<TablePtr, TablePtr>(sched_dec, child),
                grouping_columns_(grouping_columns),
                aggregation_functions_(aggregation_functions),
                mat_stat_(mat_stat) {
                }

                virtual bool execute() {
                    //std::cout << "Execute Groupby GPU" << std::endl;
                    //const TablePtr sort(TablePtr table, const std::string& column_name, SortOrder order=ASCENDING, MaterializationStatus mat_stat=MATERIALIZE, ComputeDevice comp_dev=CPU);
                    this->result_ = BaseTable::groupby(this->getInputData(), grouping_columns_, aggregation_functions_, GPU);

                    //						std::cout << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << std::endl;
                    //						std::cout << "INPUT:" << std::endl;
                    //						this->getInputData()->print();
                    //						std::cout << "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<" << std::endl;
                    //						std::cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
                    //						std::cout << "OUTPUT:" << std::endl;
                    //						this->result_->print();
                    //						std::cout << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
                    if (this->result_) {
                        setResultSize(((TablePtr) this->result_)->getNumberofRows());
                        return true;
                    } else
                        return false;
                }

                virtual ~GPU_Groupby_Operator() {
                }
            private:
                std::list<std::string> grouping_columns_;
                std::list<ColumnAggregation> aggregation_functions_;
                MaterializationStatus mat_stat_;
            };

            Physical_Operator_Map_Ptr map_init_function_groupby_operator();
			TypedOperatorPtr create_CPU_Groupby_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision&, TypedOperatorPtr left_child, TypedOperatorPtr right_child);
			TypedOperatorPtr create_GPU_Groupby_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision&, TypedOperatorPtr left_child, TypedOperatorPtr right_child);

        }//end namespace physical_operator

        //extern Map_Init_Function init_function_Groupby_operator;

        //Map_Init_Function init_function_Groupby_operator=physical_operator::map_init_function_groupby_operator; //boost::bind();

        namespace logical_operator {

			class Logical_Groupby : public hype::queryprocessing::TypedNode_Impl<TablePtr,physical_operator::map_init_function_groupby_operator> //init_function_Groupby_operator>
            {
            public:

                Logical_Groupby(const std::list<std::string>& grouping_columns,
                        const std::list<ColumnAggregation>& aggregation_functions,
                        MaterializationStatus mat_stat = MATERIALIZE,
                        hype::DeviceConstraint dev_constr = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint())
                : TypedNode_Impl<TablePtr, physical_operator::map_init_function_groupby_operator>(false, dev_constr),
                grouping_columns_(grouping_columns),
                aggregation_functions_(aggregation_functions),
                mat_stat_(mat_stat) {
                }

                virtual unsigned int getOutputResultSize() const {
                    return 10;
                }

                virtual double getCalculatedSelectivity() const {
                    return 0.3;
                }

                virtual std::string getOperationName() const {
                    return "GROUPBY";
                }
                std::string toString(bool verbose) const{
                    std::string result="GROUPBY";
                    if(verbose){
                        result+=" (";
                        std::list<std::string>::const_iterator cit;
                        for(cit=grouping_columns_.begin();cit!=grouping_columns_.end();++cit){
                            result+=*cit;
                            if(cit!=--grouping_columns_.end())
                                result+=",";
                        }
                        result+=")";
                        result+=" USING (";
                        std::list<ColumnAggregation>::const_iterator agg_func_cit;
                        for(agg_func_cit=aggregation_functions_.begin();agg_func_cit!=aggregation_functions_.end();++agg_func_cit){
                            result+=CoGaDB::util::getName(agg_func_cit->second);
                            result+="(";
                            result+=agg_func_cit->first;
                            result+=")";
                        }
                        result+=")";
                    }
                    return result;

                }
                const std::list<std::string>& getGroupingColumns() {
                    return grouping_columns_;
                }

                const std::list<ColumnAggregation>& getColumnAggregationFunctions() {
                    return aggregation_functions_;
                }

                const MaterializationStatus& getMaterializationStatus() const {
                    return mat_stat_;
                }
            private:
                std::list<std::string> grouping_columns_;
                std::list<ColumnAggregation> aggregation_functions_;
                MaterializationStatus mat_stat_;
            };

        }//end namespace logical_operator

    }//end namespace query_processing

}; //end namespace CogaDB
