
#include <core/column.hpp>
#include <hardware_optimizations/simd_acceleration.hpp>

//#define ENABLE_SIMD_ACCELERATION

namespace CoGaDB{

    using namespace std;
    
        template<class T>
        const PositionListPtr Column<T>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
            
            T value;
            if(value_for_comparison.type()!=typeid(T)){ 
                //catch some special cases
                if(typeid(T)==typeid(float) && value_for_comparison.type()==typeid(int)){
                    value = boost::any_cast<int>(value_for_comparison);
                }else{
                    std::cout << "Fatal Error!!! Typemismatch for column " << this->name_ << std::endl;
                    std::cout << "Column Type: " << typeid(T).name() << " filter value type: " << value_for_comparison.type().name() << std::endl;
                    std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                    exit(-1);
                }
            }else{
                //everything fine, filter value matches type of column
                value = boost::any_cast<T>(value_for_comparison);
            }

            PositionListPtr result_tids = PositionListPtr(new PositionList());
             unsigned int array_size=this->size();
            //calls new internally   
             result_tids->resize(array_size);
             //tids.reserve(array_size);
             //get pointer
             unsigned int* array_tids=hype::util::begin_ptr(*result_tids);
             assert(array_tids!=NULL);
             unsigned int pos=0;

            if(!quiet) std::cout << "Using CPU for Selection..." << std::endl;
            //unsigned int array_size = this->size();
            //for(TID i=0;i<array_size;i++){

                    //boost::any value = column->get(i);
                    //val = values_[i];
            T* array=hype::util::begin_ptr(values_);
            
            if(comp==EQUAL){
                for(TID i=0;i<array_size;++i){
                    if(value==array[i]){
                            //result_tids->push_back(i);
                            array_tids[pos++]=i;
                    }
//                    array_tids[pos]=i;
//                    int tmp=(value==array[i]);
//                    pos+=tmp;
                }
            }else if(comp==LESSER){
                for(TID i=0;i<array_size;++i){
                    if(array[i]<value){
                            //result_table->insert(this->fetchTuple(i));
                            //result_tids->push_back(i);
                            array_tids[pos++]=i;
                    }
//                    array_tids[pos]=i;
//                    int tmp=(value<array[i]);
//                    pos+=tmp;
                }
             }else if(comp==LESSER_EQUAL){
                for(TID i=0;i<array_size;++i){
                    if(array[i]<=value){
                            //result_tids->push_back(i);
                            //result_table->insert(this->fetchTuple(i));
                            array_tids[pos++]=i;
                    }
                } 
            }else if(comp==GREATER){
                for(TID i=0;i<array_size;++i){
                    if(array[i]>value){
                            //result_tids->push_back(i);
                            //result_table->insert(this->fetchTuple(i));
                            array_tids[pos++]=i;
                    }
                }    
            }else if(comp==GREATER_EQUAL){
                for(TID i=0;i<array_size;++i){
                    if(array[i]>=value){
                            //result_tids->push_back(i);
                            //result_table->insert(this->fetchTuple(i));
                            array_tids[pos++]=i;
                    }
                }      
            }else{

            }
            //shrink to actual result size
            result_tids->resize(pos);
            return result_tids;
        }

        
        #ifdef ENABLE_SIMD_ACCELERATION
        template<>
        inline const PositionListPtr Column<int>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
            //std::cout << "SIMD SCAN" << std::endl;
            int value;
            
            if(value_for_comparison.type()!=typeid(int)){ 
                    COGADB_FATAL_ERROR(std::string("Typemismatch for column")+this->name_
                                       +std::string(" Column Type: ") + typeid(int).name()
                                       +std::string(" filter value type: ") + value_for_comparison.type().name(),""); 
            }else{
                //everything fine, filter value matches type of column
                value = boost::any_cast<int>(value_for_comparison);
            }
            unsigned int array_size = this->size();
            int* array=hype::util::begin_ptr(values_);          
            
            return CoGaDB::simd_selection_int(array,array_size,value, comp);
            
//            
//            PositionListPtr result_tids(new PositionList());
//            //PositionListPtr result_tids(new PositionList(array_size));
//            //result_tids->reserve(array_size);
////            //ensure that at least one element is in output buffer, so we can get a pointer to it
//            //result_tids->push_back(0);
//            //get the pointer to the output buffer
//            //unsigned int* result_array = hype::util::begin_ptr(*result_tids); //new int[array_size];  
//            unsigned int* result_array = new unsigned int[array_size];  
//            assert(result_array!=NULL);
//            
//            unsigned int result_size=0;
//            int comparison_value=value;
//            if(comp==EQUAL){
//                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmpeq_epi32,==,result_size);
//            }else if(comp==LESSER){
//                //COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmplt_epi32,<,result_size);
//                    __m128i* sse_array = reinterpret_cast<__m128i*>(array); 
//                    assert(sse_array!=NULL); 
//                    int alignment_offset = ((intptr_t)sse_array)%sizeof(__m128i);
//                    const int sse_array_length = (array_size-alignment_offset)*sizeof(int)/sizeof(__m128i); 
//                    if(SIMD_DEBUG_MODE) std::cout << "SSE Array Length: " << sse_array_length << std::endl; 
//                    char* tmp_array = (char*) sse_array; 
//                    tmp_array+=alignment_offset; 
//                    sse_array=reinterpret_cast<__m128i*>(tmp_array); 
//                    if(SIMD_DEBUG_MODE)  std::cout << "array adress: "  << (void*)array <<  "sse array: " << (void*)sse_array << std::endl; 
//                    if(SIMD_DEBUG_MODE)  std::cout << "First SSE Array Element: " << ((int*)sse_array)[0] << std::endl; 
//                    unsigned int pos=0; 
//                    __m128i comp_val=_mm_set1_epi32(comparison_value); 
//                    __m128i read_value=_mm_set1_epi32 (0); 
//                    if(SIMD_DEBUG_MODE)  std::cout << "alignment_offset " << alignment_offset << std::endl; 
//                    if(alignment_offset!=0){ 
//                        if(SIMD_DEBUG_MODE) std::cout << "process first unaligned data chunk: index 0 to " << alignment_offset << std::endl; 
//                        for(unsigned int i=0;i<alignment_offset/sizeof(int);i++){ 
//                            if(SIMD_DEBUG_MODE){ 
//                                std::cout << "index "<< i << std::endl; 
//                                std::cout << "value " << array[i] << " match:" << (array[i] < comparison_value) << std::endl; 
//                            } 
//                            if(array[i] < comparison_value){ 
//                                result_array[pos++]=i; 
//                            } 
//                        } 
//                    } 
//                    for(unsigned int i=0;i<sse_array_length;i++){ 
//                            assert(((intptr_t)sse_array)%sizeof(__m128i)==0);
//                            read_value=_mm_load_si128(&sse_array[i]); 
//                            if(SIMD_DEBUG_MODE){ 
//                                std::cout << "index: " << i << std::endl; 
//                            } 
//                            __m128 comp_result = (__m128) _mm_cmplt_epi32(read_value,comp_val); 
//                            int mask= _mm_movemask_ps(comp_result); \
//                            if(SIMD_DEBUG_MODE) std::cout << "Mask: " << std::hex << mask << std::dec << std::endl; 
//                            if(mask){ 
//                                if(SIMD_DEBUG_MODE) std::cout << "at least one match!" << std::endl; 
//                                for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ 
//                                    if(SIMD_DEBUG_MODE) std::cout << "sub index: " << j << " value: " << ((mask >> j) & 1) << std::endl; 
//                                    int tmp =((mask >> j) & 1); 
//                                    result_array[pos]=i*(sizeof(__m128i)/sizeof(int))+j+(alignment_offset/sizeof(int)); 
//                                    pos+=tmp; 
//                                } 
//                            } 
//                    } 
//                    if(SIMD_DEBUG_MODE) std::cout << "Remaining offsets: " << (sse_array_length*sizeof(__m128i)/sizeof(int))+alignment_offset  << " to " << array_size << std::endl; 
//                    for(unsigned int i=(sse_array_length*sizeof(__m128i)/sizeof(int))+(alignment_offset/sizeof(int));i<array_size;i++){ 
//                        if(array[i] < comparison_value){ 
//                            result_array[pos++]=i; 
//                        } 
//                    } 
//                    result_size=pos;
//             }else if(comp==LESSER_EQUAL){
//                //add one to comparison value to compare for less equal
//                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value+1,_mm_cmplt_epi32,<,result_size);
//            }else if(comp==GREATER){
//                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmpgt_epi32,>,result_size);
//            }else if(comp==GREATER_EQUAL){
//                //substract one of comparison value to compare for greater equal
//                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value-1,_mm_cmpgt_epi32,>,result_size);
//            }else{
//                COGADB_FATAL_ERROR("Invalid ValueComparator!","");
//            }            
//            
//             
//            
//             //result_tids->resize(result_size);
//             result_tids->insert(result_tids->end(),result_array,result_array+result_size);
//             
//             delete result_array;
//             return result_tids;
        }
#endif           
        
#ifdef ENABLE_SIMD_ACCELERATION
        template<>
        inline const PositionListPtr Column<float>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
            //std::cout << "SIMD SCAN" << std::endl;
            float value;
            
            if(value_for_comparison.type()!=typeid(float)){ 
					 //allow comparison with itnegers as well
                if(value_for_comparison.type()==typeid(int)){
                    value = boost::any_cast<int>(value_for_comparison);
                }else{
                    COGADB_FATAL_ERROR(std::string("Typemismatch for column")+this->name_
                                       +std::string(" Column Type: ") + typeid(float).name()
                                       +std::string(" filter value type: ") + value_for_comparison.type().name(),""); 
					 }
            }else{
                //everything fine, filter value matches type of column
                value = boost::any_cast<float>(value_for_comparison);
            }
            unsigned int array_size = this->size();
            float* array=hype::util::begin_ptr(values_);       
            
            return CoGaDB::simd_selection_float(array,array_size,value, comp);
            
//            PositionListPtr result_tids(new PositionList());
//            //result_tids->reserve(array_size);
////            //ensure that at least one element is in output buffer, so we can get a pointer to it
//            //result_tids->push_back(0);
//            //get the pointer to the output buffer
//            unsigned int* result_array = new unsigned int[array_size];  //hype::util::begin_ptr(*result_tids); //new int[array_size];  
//            assert(result_array!=NULL);
//            
//            unsigned int result_size=0;
//            float comparison_value=value;
//            if(comp==EQUAL){
//                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpeq_ps,==,result_size);
//            }else if(comp==LESSER){
//                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmplt_ps,<,result_size);
//             }else if(comp==LESSER_EQUAL){
//                //add one to comparison value to compare for less equal
//                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmple_ps,<=,result_size);
//            }else if(comp==GREATER){
//                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpgt_ps,>,result_size);
//            }else if(comp==GREATER_EQUAL){
//                //substract one of comparison value to compare for greater equal
//                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpge_ps,>=,result_size);
//            }else{
//                COGADB_FATAL_ERROR("Invalid ValueComparator!","");
//            }            
//            
//            /**********************************************************************************/
//            
// 
//
//            /**********************************************************************************/
//            
//             result_tids->insert(result_tids->end(),result_array,result_array+result_size);
//             
//             delete result_array;
//            
//             //result_tids->resize(result_size);
//             return result_tids;
        }
#endif               
       
        template <typename T>
        void ColumnBaseTyped<T>::hash_join_pruning_thread(ColumnBaseTyped<T>* join_column, HashTable* hashtable, unsigned int* join_tids_table1, unsigned int* join_tids_table2, unsigned int thread_id, unsigned int number_of_threads, unsigned int* result_size){
            ColumnBaseTyped<T>& join_column_ref = dynamic_cast< ColumnBaseTyped<T>& >(*join_column);
            //number of elements per thread
            unsigned int chunk_size = join_column_ref.size()/number_of_threads;
            unsigned int begin_index=chunk_size*thread_id;
            unsigned int end_index;
            if(thread_id+1 == number_of_threads){
                //process until end of input array
                end_index=join_column_ref.size();
            }else{
                end_index=(chunk_size)*(thread_id+1);
            }
            if(!quiet && verbose && debug)
                std::cout << "Parallel Hash Join for array of size: " << join_column_ref.size() << " thread id: " << thread_id << " chunksize: " << chunk_size << " elements  begin index: " << begin_index << " end index: " << end_index << std::endl;
            unsigned int pos1 = begin_index;
            unsigned int pos2 = begin_index;   
            
            std::pair<typename HashTable::iterator, typename HashTable::iterator> range;
            typename HashTable::iterator it;
            for(unsigned int i=begin_index;i<end_index;i++){
                    range =  hashtable->equal_range(join_column_ref[i]);
                    for(it=range.first ; it!=range.second;++it){
                            if(it->first==join_column_ref[i]){ //(*join_column)[i]){
                                      join_tids_table1[pos1++]=it->second;
                                      join_tids_table2[pos2++]=i;
                                      if(!quiet && verbose && debug)
                                        std::cout << "match! " << it->second << ", " << i << "	"  << it->first << std::endl;
                            }
                    }
            }
            assert(pos1==pos2);
            *result_size=pos1;
        }
        
        
	template<class T>
	const PositionListPairPtr ColumnBaseTyped<T>::parallel_hash_join(ColumnPtr join_column_,unsigned int number_of_threads){

		

                if(join_column_->type()!=typeid(T)){
                        std::cout << "Fatal Error!!! Typemismatch for columns " << this->name_  << " and " << join_column_->getName() << std::endl;
                        std::cout << "File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                        exit(-1);
                }

                shared_pointer_namespace::shared_ptr<ColumnBaseTyped<T> > join_column = shared_pointer_namespace::static_pointer_cast<ColumnBaseTyped<T> >(join_column_); //static_cast<IntColumnPtr>(column1);

                PositionListPairPtr join_tids( new PositionListPair());
                join_tids->first = PositionListPtr( new PositionList() );
                join_tids->second = PositionListPtr( new PositionList() );

                Timestamp build_hashtable_begin = getTimestamp();
               //create hash table
               HashTable hashtable;
               unsigned int hash_table_size=this->size();
               unsigned int join_column_size=join_column->size();

               assert(join_column_->size()>=this->size());                        
               unsigned int* join_tids_table1 =  new unsigned int[join_column_size];   
               unsigned int* join_tids_table2 =  new unsigned int[join_column_size]; 

//               ColumnBaseTyped<T>& join_column_ref = dynamic_cast< ColumnBaseTyped<T>& >(*join_column);

                for(unsigned int i=0;i<hash_table_size;i++)	
                        hashtable.insert(std::pair<T,TID> ((*this)[i],i));
                Timestamp build_hashtable_end = getTimestamp();

                //probe larger relation
                Timestamp prune_hashtable_begin = getTimestamp();
                std::vector<unsigned int> thread_result_sizes(number_of_threads);
		boost::thread_group threads;
		for(unsigned int i=0;i<number_of_threads;i++){
                    //ColumnBaseTyped<T>::hash_join_pruning_thread(join_column.get(), &hashtable, join_tids_table1, join_tids_table2, i, number_of_threads);
                    //CoGaDB::hash_join_pruning_thread<T,HashTable>(join_column_ref, &hashtable, join_tids_table1, join_tids_table2, i, number_of_threads);
                    //threads.add_thread(new boost::thread((&hash_join_pruning_thread<T,HashTable>),join_column_ref, &hashtable, join_tids_table1, join_tids_table2, i, number_of_threads));
                    //ColumnBaseTyped<T>& join_column_ref, HashTable* hashtable, unsigned int* join_tids_table1, unsigned int* join_tids_table2, unsigned int thread_id, unsigned int number_of_threads)
                    threads.add_thread(new boost::thread(boost::bind(&ColumnBaseTyped<T>::hash_join_pruning_thread, join_column.get(), &hashtable, join_tids_table1, join_tids_table2, i, number_of_threads, &thread_result_sizes[i])));
		}
		threads.join_all();
                
                for(unsigned int i=0;i<number_of_threads;i++){
                    
                    cout << "thread: " << i << " result size: " << thread_result_sizes[i] << endl;
                }
                
                //put results in position list
                join_tids->first->insert(join_tids->first->end(),join_tids_table1,join_tids_table1+join_column_size); //assume PK FK Join!
                join_tids->second->insert(join_tids->second->end(),join_tids_table2,join_tids_table2+join_column_size);
                //cleanup
                delete join_tids_table1;
                delete join_tids_table2;
                
                Timestamp prune_hashtable_end = getTimestamp();

                if(!quiet && verbose && debug)
                std::cout << "Hash Join: Build Phase: " << double(build_hashtable_end-build_hashtable_begin)/(1000*1000) << "ms"
                          << "Pruning Phase: "  << double(prune_hashtable_end-prune_hashtable_begin)/(1000*1000) << "ms" << std::endl; 

		return join_tids;
	}
        
        
        void column_template_instatiation(){
                Column<int> col_int("",INT);
                Column<float> col_float("",FLOAT);
                Column<std::string> col_string("",VARCHAR);
        }

}; //end namespace CoGaDB
        
        