
#include <core/gpu_column_cache.hpp>

namespace CoGaDB{
using namespace std;

		GPU_Column_Cache::GPU_Column_Cache(unsigned int max_gpu_buffer_size_in_byte) : map_(),
																											  max_gpu_buffer_size_in_byte_(max_gpu_buffer_size_in_byte), 
																											  caching_enabled_(true){

		}

		GPU_Column_Cache& GPU_Column_Cache::instance(){
			static GPU_Column_Cache cache(getTotalGPUMemorySizeInByte()*0.5); //use half of the total device memory as buffer for columns 
			return cache;
		}	

		void GPU_Column_Cache::printStatus() const throw(){
			cout << "GPU_Column_Cache Status:" << endl;
			cout << "GPU Buffer Size: " << max_gpu_buffer_size_in_byte_/(1024*1024) << "MB" << endl; 
			cout << "Total GPU Memory: " << getTotalGPUMemorySizeInByte()/(1024*1024) << "MB" << endl; 
			cout << "Free GPU Memory: " << getFreeGPUMemorySizeInByte()/(1024*1024) << "MB" << endl; 
			cout << "Cached columns:" << endl;
			Map::const_iterator cit;
			for(cit=map_.begin();cit!=map_.end();++cit){
				cout << cit->second->getName() << " of Size " << cit->second->size() << " elements, currently in use: " << !cit->second.unique() << endl;
			}
		}

		const gpu::GPU_Base_ColumnPtr GPU_Column_Cache::getGPUColumn(ColumnPtr column_ptr){
			gpu::GPU_Base_ColumnPtr result_ptr;
			if(!column_ptr) return gpu::GPU_Base_ColumnPtr();
			if(!caching_enabled_){
				result_ptr=gpu::copy_column_host_to_device(column_ptr);
				return result_ptr;
			}

			Map::iterator it=map_.find(column_ptr);
			if(it==map_.end()){
				cout << "Cache Miss: " << column_ptr->getName() << endl;

				unsigned int used_memory_in_bytes=getTotalGPUMemorySizeInByte()-getFreeGPUMemorySizeInByte();

				if(column_ptr->getSizeinBytes()+used_memory_in_bytes>max_gpu_buffer_size_in_byte_){
					cout << "Warning: GPU_Column_Cache::getGPUColumn(): Unsufficient Device Memory" << endl;
					//apply column replacement strategy!
					return gpu::GPU_Base_ColumnPtr();
				}

				result_ptr=gpu::copy_column_host_to_device(column_ptr);
				if(!result_ptr){ 
					cout << "Warning: GPU_Column_Cache::getGPUColumn(): Unsufficient Device Memory" << endl;
					return gpu::GPU_Base_ColumnPtr();
				}		
				//if(caching_enabled_){
					map_.insert( std::make_pair(column_ptr, result_ptr));
				//}
				return result_ptr;
			}else{
				cout << "Cache Hit: " << column_ptr->getName() << endl;
				return it->second; //return pointer to GPU memory
			}

		}


		bool GPU_Column_Cache::removeGPUColumn(gpu::GPU_Base_ColumnPtr dev_col){
			Map::iterator it;
			for(it=map_.begin();it!=map_.end();++it){
				if(it->second==dev_col){
					map_.erase(it);
					return true;
				}
			}
			return false;
		}


}; //end namespace CoGaDB
