
#include <core/table.hpp>
#include <core/runtime_configuration.hpp>
#include <iostream>
#include <fstream>
#include <boost/filesystem.hpp>
#include <boost/any.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>

#include <util/filesystem.hpp>
#include <util/time_measurement.hpp>
#include <util/iostream.hpp>

namespace CoGaDB{

	using namespace std;

//	const TableSchema mergeTableSchemas(const TableSchema& schema1, const std::string& join_attributname_table1,
//													const TableSchema& schema2, const std::string& join_attributname_table2){
// 		 TableSchema result_schema, schema_(schema1), schema(schema2);
//		 TableSchema::iterator it,it2;
//		 map<string, pair<AttributeType,string> > schema_map;
//		 map<string, pair<AttributeType,string> >::iterator map_it;

//		

//		 for(it=schema_.begin();it!=schema_.end();it++){
//			 schema_map[it->second]=*it;
//			 cout << "it->second: " << it->second << endl;
//			 result_schema.push_back(*it);
//		 }

//		//TableSchema schema=table->getSchema();
//		 for(it2=schema.begin();it2!=schema.end();it2++){
//			map_it=schema_map.find(it2->second);
//			if(map_it!=schema_map.end()){
//				schema_map[it2->second]=*it2;
//				cout << "FOUND Duplicate!" << it2->second << endl;
//				result_schema.push_back(*it2);	
//			}else{
//				schema_map[it2->second]=*it2;
//				cout << "it2->second: " << it2->second << endl;
//				result_schema.push_back(*it2);				
//			}
//			
//		 }
//		cout << "[DEBUG]: JOIN(): schema of new table:" << endl << "| ";
//		 for(it=result_schema.begin();it!=result_schema.end();it++)
//			cout << (it->second) << " |"; // << endl;

//		cout << endl;
//		return result_schema;
//	}


	const std::vector<ColumnPtr>& Table::getColumns() const{
		return columns_;
	}

	Table::Table(const std::string& name, const TableSchema& schema) : BaseTable(name,schema), columns_(){
		TableSchema::const_iterator it;
		for(it=schema.begin();it!=schema.end();it++){
			ColumnPtr ptr = createColumn(it->first,it->second);
			if(ptr.get()!=NULL){
				columns_.push_back(ptr);
			}
		}
	}

	Table::Table(const std::string& name, const std::vector<ColumnPtr>& columns) : BaseTable(name,TableSchema()), columns_(columns){
		vector<ColumnPtr>::const_iterator it;
		for(it=columns.begin();it!=columns.end();it++){
			schema_.push_back( Attribut((*it)->getType(),(*it)->getName()) );
		}
	}

	Table::~Table(){

	}


	void Table::print(){
			cout << endl << name_ << ":" << endl;
			cout << "|\t";	
			for(unsigned int i=0;i<columns_.size();i++){
				cout << columns_[i]->getName() << "\t|\t";
			}
			cout << endl;
			cout << "-------------------------------" << endl;
			if(columns_.empty()){
				cout << "Table " << name_ << " is Empty, nothing to print" << endl;
				return;
			}
			for(unsigned int j=0;j<columns_[0]->size();j++){
				cout << "|\t";
				for(unsigned int i=0;i<columns_.size();i++){
					boost::any value = columns_[i]->get(j);
					if(value.type()==typeid(string)){
						cout << boost::any_cast<string>(value) << "\t|\t"; 
					}else if (value.type()==typeid(int)){
						cout << boost::any_cast<int>(value) << "\t|\t"; 
					}else if (value.type()==typeid(float)){
						cout << setiosflags(ios::fixed) << boost::any_cast<float>(value) << "\t|\t"; 
					}else if (value.type()==typeid(bool)){
						cout << boost::any_cast<bool>(value) << "\t|\t"; 
					}else{

					}


					
				}
				cout << endl;
			}
			cout << endl << columns_[0]->size() << " rows" << endl << endl;
	}

	bool Table::store(){

		using namespace boost::filesystem;
		string dir_path(RuntimeConfiguration::instance().getPathToDatabase());
		if ( !exists( dir_path ) ) return false;



		dir_path += "/tables/";
		dir_path += name_;		
		if ( !exists( dir_path ) ){
			create_directory(dir_path);
		}
		//store all files (colums)
		for(unsigned int i=0;i<columns_.size();i++){
			columns_[i]->store(dir_path);
		}

		return true;
	}

	bool Table::load(){

		using namespace boost::filesystem;
		string dir_path(RuntimeConfiguration::instance().getPathToDatabase());
		if ( !exists( dir_path ) ) return false;
		dir_path += "/tables/";
		dir_path += name_;
		if ( !exists( dir_path ) ){
			cout << "No directory '" << dir_path  << "' Aborting..." << endl;
			return false;
		}

		vector<string> v = getFilesinDirectory(dir_path);
		//for(unsigned int i=0;i<v.size();i++) cout << v[i] <<endl;

		boost::thread_group threads;

		for(unsigned int i=0;i<columns_.size();i++){
			//columns_[i]->load(dir_path);
			//boost::thread* thr = new boost::thread(boost::bind(&Foo::some_function, &f));
			threads.add_thread(new boost::thread(boost::bind(&ColumnBase::load, columns_[i],dir_path)));
		}
		threads.join_all();
		
//		for(unsigned int i=0;i<columns_.size();i++){
//			columns_[i]->load(dir_path);
//		}




		return true;
	}
	bool Table::loadDatafromFile(std::string filepath){

		std::ifstream fin(filepath.c_str());  //"Makefile");
		std::string buffer;

 		std::cout << "Hier der Inhalt der Datei:\n";

		unsigned int samplecounter=0;
		if(!fin.is_open()){
			cout << "Error: could not open file " << filepath << endl;
			return false;
		}
		while (fin.good()) {
			getline(fin,buffer,'\n');
			//std::cout <<buffer<< std::endl;

			std::string s = buffer;
			std::string str = ";;Hello|world||-foo--bar;yow;baz|";
			 str = buffer; //s;

			std::string xvalue=""; 
			std::string yvalue=""; 

		  typedef boost::tokenizer<boost::char_separator<char> >  tokenizer;
			 boost::char_separator<char> sep("|");
			 tokenizer tokens(str, sep);

			unsigned int tokencounter=0;

			//create tuple from string
			Tuple t;
			for (tokenizer::iterator tok_iter = tokens.begin(); tok_iter != tokens.end(); ++tok_iter){	
				tokenizer::iterator tmp_tok_it=tok_iter;
				tmp_tok_it++;
				if(tmp_tok_it == tokens.end() && (*tok_iter == "" || *tok_iter == "\r")) break; //if reached last token and token is null value, then process next line

				if(tokencounter>=columns_.size()){
					cout << "Fatal Error! While parsing file '" << filepath << "' by processing line: " << samplecounter << " Token: " << tokencounter << endl;
					cout << "More Tokens found than columns in Table '" << name_ << "'" << endl;
					return false;
				} 
				AttributeType type = columns_[tokencounter]->getType();
				if(type==INT){
					t.push_back(boost::lexical_cast<int>(*tok_iter));
				}else if(type==FLOAT){
					t.push_back(boost::lexical_cast<float>(*tok_iter));
				}else if(type==VARCHAR){
					t.push_back(*tok_iter); //is already a string
				}else if(type==BOOLEAN){
					t.push_back(boost::lexical_cast<bool>(*tok_iter));
				}


				tokencounter++;
			}

			//cout << "Processed Line: '" << str << "'" << endl;
			if(str!="") //if processed line is empty, the tuple is empty as well, so to ignore empty lines, we just do not insert tuples coming from empty lines 
			if(this->insert(t)==false){
				cout << "Failed to load data from file: '" << filepath << "'" << endl;
				return false;
			}
			if(samplecounter%10000==0) cout << "Processed " << samplecounter << " rows..." << endl;
			samplecounter++;
		}
		fin.close();
		
		return true;
	}

	const TablePtr Table::materialize() const{

		return TablePtr();
	}
        
        bool Table::addColumn(ColumnPtr col){
            if(!col) return false;
            this->columns_.push_back(col);
            this->schema_.push_back(Attribut(col->getType(),col->getName()));
            return true;
        }

	/***************** status report *****************/
	unsigned int Table::getNumberofRows() const throw(){
		if(columns_.size()>0){
			return columns_[0]->size();
		}else{
			return 0;
		}
	}

	bool Table::isMaterialized() const throw(){
		return true;
	}

	
	/***************** relational operations *****************/
	/*
	const TablePtr Table::selection(const std::string& column_name, const boost::any& value_for_comparison, const ValueComparator& comp, const ComputeDevice& comp_dev) const{
			Timestamp begin,end; 
			begin=getTimestamp();
			ColumnPtr column = this->getColumnbyName(column_name);
			if(column.get()==NULL){
				cout << "Error! Could not look up Column: " << column_name << endl;
				exit(-1);
			}
			
			TablePtr result_table(new Table(string("tmp_selection_on_")+column_name,schema_)) ;//tmp_schema));


		//perform selection
		vector<TID> tids=column->selection(value_for_comparison, comp, comp_dev);

		//fetch result tuples
		for(unsigned int i=0;i<tids.size();i++){
			TID j = tids[i];
			result_table->insert( this->fetchTuple(j) );
		}

		end=getTimestamp();
		assert(end>=begin);
		cout << "Time for Selection: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;
		return result_table;
	}

	const TablePtr Table::projection(const std::list<std::string>& columns_to_select, const ComputeDevice comp_dev) const {

	Timestamp begin,end; 
		begin=getTimestamp();

		list<string>::const_iterator it;
		vector<ColumnPtr> new_schema;
		for(it=columns_to_select.begin(); it!=columns_to_select.end(); it++){
			ColumnPtr column = this->getColumnbyName(*it);
			if(column.get()==NULL){
				cout << "Error! " << endl;
				exit(-1);
			}
			ColumnPtr new_column = column->copy();
			new_schema.push_back(new_column);
		}

		
		TablePtr result_table(new Table(string("tmp_projection_on_table_")+name_,new_schema)) ;//tmp_schema));
		//TablePtr result_table(new Table(string(""),new_schema)) ;//tmp_schema));

		end=getTimestamp();
		assert(end>=begin);
		cout << "Time for Projection: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;

		return result_table;
	}

	const TablePtr Table::join(TablePtr table_, const std::string& join_column_table1, const std::string& join_column_table2, const ComputeDevice comp_dev) const{

		shared_pointer_namespace::shared_ptr<Table> table = shared_pointer_namespace::static_pointer_cast<Table>(table_);

		ColumnPtr col1 = this->getColumnbyName(join_column_table1);
		if(col1==NULL){
			cout << "table " << this->getName() << " has no Column named " << join_column_table1 << endl;
			cout << "join failed!" << endl;
			return TablePtr();
		}
		ColumnPtr col2 = this->getColumnbyName(join_column_table2);
		if(col2==NULL){
			cout << "table " << table->getName() << " has no Column named " << join_column_table2 << endl;
			cout << "join failed!" << endl;
			return TablePtr();
		}

		std::vector<TID_Pair> join_tids = col1->sort_merge_join(col2,comp_dev);

		TableSchema result_schema;
		result_schema=mergeTableSchemas(schema_,join_column_table1,table->getSchema(),join_column_table2);

//materialize Table
		TablePtr result_table(new Table(string("tmp_join_of_")+name_+string(".")+join_column_table1
																	   +string(",")+table->getName()+string(".")+join_column_table2,result_schema));
			TID index_table1,index_table2;
			for(unsigned int i=0;i<join_tids.size();i++){
				index_table1 = join_tids[i].first;
				index_table2 = join_tids[i].second;

				Tuple t;
				for(unsigned int i=0;i<columns_.size();i++)
					t.push_back( columns_[i]->get(index_table1) );

				for(unsigned int i=0;i<table->columns_.size();i++)
					t.push_back( table->columns_[i]->get(index_table2) );				

				result_table->insert(t);				
			}

		return result_table;
	}

	const TablePtr Table::sort(const std::string& column_name, SortOrder order, ComputeDevice comp_dev) const{
Timestamp begin,end; 
		begin=getTimestamp();
		ColumnPtr column = this->getColumnbyName(column_name);
			if(column.get()==NULL){
				cout << "Error! Could not look up Column: " << column_name << endl;
				exit(-1);
			}
			
			vector<TID> ids = column->sort(comp_dev);

			TablePtr result_table(new Table(string("tmp_sort_on_")+name_+string(".")+column_name,schema_));


			//cout << "adding tuples to new table " << endl;
			for(unsigned int i=0;i<column->size();i++){
				TID j = ids[i];
				result_table->insert( this->fetchTuple(j) );
			}
		end=getTimestamp();
		assert(end>=begin);
		cout << "Time for Sorting: " << end-begin << "ns (" << double(end-begin)/1000000 << "ms)" << endl;
		return result_table;
	}

	const TablePtr Table::groupby(const std::string& grouping_column, const std::string& aggregation_column,  AggregationMethod agg_meth, ComputeDevice comp_dev) const {

		TableSchema schema;
		schema.push_back( Attribut(VARCHAR,grouping_column) );
		schema.push_back( Attribut(INT,aggregation_column) );
		TablePtr result_table(new Table(string("tmp_groupby_of_")+name_+string(".")+grouping_column,schema));
		

		ColumnPtr column = this->getColumnbyName(grouping_column);
		if(column.get()==NULL){
			cout << "Error! Could not look up Column: " << column << endl;
			exit(-1);
		}
		ColumnPtr column_agg = this->getColumnbyName(aggregation_column);
		if(column_agg.get()==NULL){
			cout << "Error! Could not look up Column: " << aggregation_column << endl;
			exit(-1);
		}
			
			
		vector<TID> ids = column->sort(comp_dev);

		TID begin_index=0, end_index=0; //1;//0;//1;

		cout << "Grouping column:" << column->getName() << " Aggregation Column: " << column_agg->getName() << endl;

			for(TID i=0;i<ids.size();i++){
				cout << "Value: " << boost::any_cast<string>(column->get(ids[i])) << " AGGREGATE: " << boost::any_cast<int>(column_agg->get(ids[i])) << " ID: " << i <<  endl;

			}

		while(end_index<ids.size()){
			boost::any value = column->get(ids[begin_index]);
			boost::any value2 = column->get(ids[end_index]);
			string val1 = boost::any_cast<string>(value);
			string val2 = boost::any_cast<string>(value2);
			end_index++;
			while( end_index<ids.size() && boost::any_cast<string>( column->get(ids[begin_index]) )==boost::any_cast<string>( column->get(ids[end_index]) )){
				
				end_index++;
				cout << "index begin: " << begin_index << " index end: " << end_index << " index begin real: " << ids[begin_index] << " index end real: " << ids[end_index] << "Size: " << ids.size()  << endl;
				cout << "start entry: " << boost::any_cast<string>( column->get(ids[begin_index]) ) << " end entry:" << boost::any_cast<string>( column->get(ids[end_index-1]) ) << endl;
			}
			//if(end_index-begin_index>2)
			//end_index++;
			
			//aggregate!!!
			int result=0;
			cout << "aggregate values for entry: " << boost::any_cast<string>( column->get(ids[begin_index]) ) << endl;
			for(TID i=begin_index;i<end_index;i++){
				cout << "\tindex begin: " << begin_index << " index end: " << end_index << " current index: " << i << " real TID: " << ids[i] << endl;
				cout << "\tValue: " << boost::any_cast<string>(column->get(ids[i])) << " AGGREGATE: " << boost::any_cast<int>(column_agg->get(ids[i])) << endl;
				result += boost::any_cast<int>(column_agg->get(ids[i]));

			}
			Tuple t;
			t.push_back(boost::any_cast<string>( column->get(ids[begin_index]) ));
			t.push_back(result);

			result_table->insert(t);
			//end_index++;
			begin_index=end_index; //-1;

		}

		return result_table;
	}
	*/
	/***************** read and write operations at table level *****************/
	const Tuple Table::fetchTuple(const TID& id) const{
		//TODO: add check the bounding and whether id is valid
		Tuple t;
		for(unsigned int j=0;j<columns_.size();j++){			
			t.push_back(columns_[j]->get(id));
		}
		return t;
	}

	bool Table::insert(const Tuple& t){
		if(t.size()!=columns_.size()){
			cout << "Fatal Error! Tuple has invalid number of values!" << endl;
			cout << "Mumber of Columns in Table '" << name_ << "': " << columns_.size() << endl;
			cout << "Number of values in Tuple: " << t.size() << endl;
			cout << t << endl;
			return false;
		}
		bool ret_val=true;
		Tuple::const_iterator it;
		unsigned int i=0;
		for(it=t.begin();it!=t.end();it++,i++){
			ret_val = columns_[i]->insert(*it);
			if(ret_val == false){
			cout << "Fatal Error! could not insert value in Column " << columns_[i]->getName() << endl; 
			return false;
			}
		}
		return true;
	}

	bool Table::update(const std::string& attribute_name, const boost::any& ){ //value){
		// TODO: add implementation
		ColumnPtr col = this->getColumnbyName(attribute_name);
		if(col){
			//return col->update();
		}
		return false;
	}

	bool Table::remove(const std::string& attribute_name, const boost::any& ){ //value){
		ColumnPtr col = this->getColumnbyName(attribute_name);
		if(col){
			//return col->update();
		}
		return false;
	}

	
	
	const ColumnPtr Table::getColumnbyName(const std::string& column_name) const throw(){
		for(unsigned int i=0;i<columns_.size();i++){
				if(!quiet && verbose) cout << "Looking up column: " << columns_[i]->getName() << " Look for Name: " << column_name << endl;
				if(columns_[i]->getName()==column_name){
					  if(!quiet && verbose) 
                                              cout << "Found!" << endl;
					  if(!quiet && verbose) 
                                              cout << columns_[i].get() << endl;
					 return columns_[i];
				}
		}
		cout << "Error: could not find column " << column_name << " in Table '" << name_ << "'!" << endl;
		return ColumnPtr(); //return NULL Pointer  //shared_ptr<ColumnBase>((ColumnBase*) 0);
	}

	/***************** relational operations that return lookup tables *****************/
	/*
	const std::vector<TID> Table::lookup_selection(const std::string& column_name, const boost::any& value_for_comparison, const ValueComparator& comp, const ComputeDevice& comp_dev) const{

		return std::vector<TID>();
	}

//	//const TablePtr Table::projection(const std::list<std::string>& columns_to_select, const ComputeDevice comp_dev) const{

//		return;
//	}

	const std::vector<CoGaDB::TID_Pair> Table::lookup_join(TablePtr table, const std::string& join_column_table1, const std::string& join_column_table2, const ComputeDevice comp_dev) const{

		return std::vector<CoGaDB::TID_Pair>();
	}

	const TablePtr Table::lookup_sort(const std::string& column_name, SortOrder order, ComputeDevice comp_dev) const{

		return TablePtr();
	}
	*/
}; //end namespace CogaDB


