#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>
#include <gpu/gpu_positionlist.hpp>
#include <core/column.hpp>

using namespace std;

namespace CoGaDB {
    namespace gpu {
        //using namespace CoGaDB::core;

        GPU_Positionlist::~GPU_Positionlist() {
        }

        GPU_Base_Column::GPU_Base_Column(const std::string& name, AttributeType db_type) : name_(name), db_type_(db_type) {

        }

        GPU_Base_Column::~GPU_Base_Column() {

        }

        /*! \brief returns database type of column (as defined in "SQL" statement)*/
        AttributeType GPU_Base_Column::getType() const throw () {
            return this->db_type_;
        }

        const std::string GPU_Base_Column::getName() const throw () {
            return this->name_;
        }

        const ColumnPtr copy_column_device_to_host(GPU_Base_ColumnPtr device_column) {

            if (!device_column)
                return ColumnPtr();

            ColumnPtr result_host_ptr;

            AttributeType type = device_column->getType();
            string name = device_column->getName();

            if (type == INT) {
                shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (device_column);
                assert(dev_col != NULL);
                thrust::device_vector<int>& dev_data = dev_col->getContent();
                Column<int>* host_col = new Column<int>(device_column->getName(), INT);
                vector<int>& host_data = host_col->getContent();
                host_data.resize(dev_data.size());
                thrust::copy(dev_data.begin(), dev_data.end(), host_data.begin());
                result_host_ptr = ColumnPtr(host_col);
            } else if (type == FLOAT) {
                shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (device_column);
                assert(dev_col != NULL);
                thrust::device_vector<float>& dev_data = dev_col->getContent();
                Column<float>* host_col = new Column<float>(device_column->getName(), FLOAT);
                vector<float>& host_data = host_col->getContent();
                host_data.resize(dev_data.size());
                thrust::copy(dev_data.begin(), dev_data.end(), host_data.begin());
                result_host_ptr = ColumnPtr(host_col);
            } else if (type == VARCHAR) {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: VARCHAR is currently not supported, will be added again in the future!" << endl;
            } else if (type == BOOLEAN) {
                //ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: bool is currently not supported, will be added again in the future!" << endl;
            } else {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
            }

            return result_host_ptr;

        }

        const GPU_Base_ColumnPtr copy_column_host_to_device(ColumnPtr host_column) {

            if (!host_column)
                return GPU_Base_ColumnPtr();

            GPU_Base_ColumnPtr result_device_ptr;



            AttributeType type = host_column->getType();
            string name = host_column->getName();

            cout << "Copy Column " << name << " of type " << type << endl;
            cout << "Size " << host_column->size() << " Materialized: " << host_column->isMaterialized() << endl;

            if (!host_column->isMaterialized()) host_column = host_column-> materialize();

            cout << "===================" << endl;
            cout << "Copy Column " << name << " of type " << type << endl;
            cout << "Size " << host_column->size() << " Materialized: " << host_column->isMaterialized() << endl;

            if (type == INT) {
                shared_pointer_namespace::shared_ptr<Column<int> > host_col = shared_pointer_namespace::static_pointer_cast<Column<int> > (host_column);
                assert(host_col != NULL);
                vector<int>& host_data = host_col->getContent();
                GPU_Typed_Base_Column<int>* dev_col = new GPU_Typed_Base_Column<int>(name, type);
                thrust::device_vector<int>& dev_data = dev_col->getContent();
                //cout << "Allocate " << host_data.size() << " int memory" << endl;
                try {
                    dev_data.resize(host_data.size());
                    thrust::copy(host_data.begin(), host_data.end(), dev_data.begin());
                    result_device_ptr = GPU_Base_ColumnPtr(dev_col);
                } catch (thrust::system::detail::bad_alloc& e) {
                    cout << e.what() << endl;
                    printGPUStatus();
                    result_device_ptr = GPU_Base_ColumnPtr(); //set to NULL Pointer
                }

            } else if (type == FLOAT) {
                shared_pointer_namespace::shared_ptr<Column<float> > host_col = shared_pointer_namespace::static_pointer_cast<Column<float> > (host_column);
                assert(host_col != NULL);
                //cout << "host_col->size(): " << host_col->size() << endl;
                vector<float>& host_data = host_col->getContent();
                GPU_Typed_Base_Column<float>* dev_col = new GPU_Typed_Base_Column<float>(name, type);
                thrust::device_vector<float>& dev_data = dev_col->getContent();
                //cout << "Allocate " << host_data.size() << " floats memory" << endl;
                try {
                    dev_data.resize(host_data.size());
                    thrust::copy(host_data.begin(), host_data.end(), dev_data.begin());
                    result_device_ptr = GPU_Base_ColumnPtr(dev_col);
                } catch (thrust::system::detail::bad_alloc& e) {
                    cout << e.what() << endl;
                    printGPUStatus();
                    result_device_ptr = GPU_Base_ColumnPtr(); //set to NULL Pointer
                }
            } else if (type == VARCHAR) {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: VARCHAR is currently not supported, will be added again in the future!" << endl;
            } else if (type == BOOLEAN) {
                //ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: bool is currently not supported, will be added again in the future!" << endl;
            } else {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
            }

            return result_device_ptr;
        }

        //		const PositionListPtr copy_PositionList_device_to_host(GPU_Positionlist device_pos_list){
        //				std::vector<TID>* host_vec_ptr = new std::vector<TID>(pos_list_.size());
        //				thrust::copy(pos_list_.begin(),pos_list_.end(),host_vec_ptr->begin());
        //				return host_vec_ptr;
        //		}


        //		const GPU_Base_ColumnPtr copy_PositionList_host_to_device(PositionListPtr pos_list){

        //		}

        const PositionListPtr copy_PositionList_device_to_host(GPU_PositionlistPtr device_pos_list) {
            //			shared_pointer_namespace::shared_ptr<Impl_GPU_Positionlist> dev_pos_list_ptr = shared_pointer_namespace::static_pointer_cast<Impl_GPU_Positionlist> (device_pos_list);
            assert(device_pos_list != NULL);
            //std::cout << "calling copy_PositionList_device_to_host" << std::endl;
            return device_pos_list->toHostPositionlist(); //PositionListPtr();

        }

        const GPU_PositionlistPtr copy_PositionList_host_to_device(PositionListPtr pos_list) {

            return GPU_PositionlistPtr(new Impl_GPU_Positionlist(pos_list.get()));
        }

        const GPU_Base_ColumnPtr createGPUColumn(AttributeType type, const std::string& name) {
            GPU_Base_ColumnPtr ptr;
            if (type == INT) {
                //			if(typeid(vector<int>&)!=host_vector.type()){
                //				cout << "Type Mismatch! Specified type " << typeid(int).name() << "for GPU_Base_Column, but passed data vector is not from type vector<int>!" << endl;
                //				cout << "Passed Object is from type " << host_vector.type().name() << endl;
                //				cout << "File: __FILE__ Line: __LINE__" << endl;
                //				exit(-1);
                //			}
                //			const vector<int>& v = boost::any_cast<vector<int>&>(host_vector);
                ptr = GPU_Base_ColumnPtr(new GPU_Typed_Base_Column<int>(name, INT));
            } else if (type == FLOAT) {
                //			if(typeid(vector<float>&)!=host_vector.type()){
                //				cout << "Type Mismatch! Specified type " << typeid(int).name() << "for GPU_Base_Column, but passed data vector is not from type vector<int>!" << endl;
                //				cout << "Passed Object is from type " << host_vector.type().name() << endl;
                //				cout << "File: __FILE__ Line: __LINE__" << endl;
                //				exit(-1);
                //			}
                //			const vector<int>& v = boost::any_cast<vector<int>&>(host_vector);
                ptr = GPU_Base_ColumnPtr(new GPU_Typed_Base_Column<float>(name, FLOAT));
            } else if (type == VARCHAR) {
                //		ptr=ColumnPtr(new Column<string>(name,VARCHAR));
                //ptr=NULL;
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: bool is currently not supported, will be added again in the future!" << endl;
            } else if (type == BOOLEAN) {
                //ptr=NULL;
                //ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
                        << " Note: bool is currently not supported, will be added again in the future!" << endl;
            } else {
                cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
            }

            return ptr;
        }

        const GPU_Base_ColumnPtr materializeColumn(GPU_Base_ColumnPtr device_column, GPU_PositionlistPtr device_pos_list) {
            return GPU_Base_ColumnPtr();
        }



        //	//boost::any is workaround, so that vectors of different types can be passed via the interface
        //	const GPU_Base_Column* createGPUColumn(AttributeType type, const std::string& name){
        //		GPU_Base_Column* ptr(NULL);
        //		if(type==INT){
        ////			if(typeid(vector<int>&)!=host_vector.type()){
        ////				cout << "Type Mismatch! Specified type " << typeid(int).name() << "for GPU_Base_Column, but passed data vector is not from type vector<int>!" << endl;
        ////				cout << "Passed Object is from type " << host_vector.type().name() << endl;
        ////				cout << "File: __FILE__ Line: __LINE__" << endl;
        ////				exit(-1);
        ////			}
        ////			const vector<int>& v = boost::any_cast<vector<int>&>(host_vector);
        //			ptr=new GPU_Typed_Base_Column<int>(name,INT);
        //		}else if(type==FLOAT){
        ////			if(typeid(vector<float>&)!=host_vector.type()){
        ////				cout << "Type Mismatch! Specified type " << typeid(int).name() << "for GPU_Base_Column, but passed data vector is not from type vector<int>!" << endl;
        ////				cout << "Passed Object is from type " << host_vector.type().name() << endl;
        ////				cout << "File: __FILE__ Line: __LINE__" << endl;
        ////				exit(-1);
        ////			}
        ////			const vector<int>& v = boost::any_cast<vector<int>&>(host_vector);
        //			ptr=new GPU_Typed_Base_Column<float>(name,FLOAT);
        //		}else if(type==VARCHAR){
        ////		ptr=ColumnPtr(new Column<string>(name,VARCHAR));
        //			ptr=NULL;
        //			cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
        //			 	  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
        //		}else if(type==BOOLEAN){
        //			ptr=NULL;
        //			//ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
        //			cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name
        //				  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
        //		}else{
        //			cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
        //		}

        //		return ptr;
        //	}

        bool copy_column_host_to_device(const std::vector<int>& host_column, GPU_Base_Column* device_column) {
            if (!device_column)
                return false;

            GPU_Typed_Base_Column<int>* d_col_ptr = static_cast<GPU_Typed_Base_Column<int>*> (device_column);
            if (!d_col_ptr)
                return false;

            return d_col_ptr->insert_from_host_memory(host_column);
        }

        bool copy_column_host_to_device(const std::vector<float>& host_column, GPU_Base_Column* device_column) {
            if (!device_column)
                return false;

            GPU_Typed_Base_Column<float>* d_col_ptr = static_cast<GPU_Typed_Base_Column<float>*> (device_column);
            if (!d_col_ptr)
                return false;

            return d_col_ptr->insert_from_host_memory(host_column);
        }
        //
        //GPU_PositionList* copy_positionlist_host_to_device(std::vector<TID>*);

        //host columns are not const, because data is inserted in real columns directly (Attention! deletes content of host_column before copying data in it!)

        bool copy_column_device_to_host(GPU_Base_Column* device_column, std::vector<int>& host_column) {
            if (!device_column)
                return false;

            GPU_Typed_Base_Column<int>* d_col_ptr = static_cast<GPU_Typed_Base_Column<int>*> (device_column);
            if (!d_col_ptr)
                return false;

            return d_col_ptr->insert_from_device_memory_in_host_vector(host_column);
        }

        //host columns are not const, because data is inserted in real columns directly (Attention! deletes content of host_column before copying data in it!)

        bool copy_column_device_to_host(GPU_Base_Column* device_column, std::vector<float>& host_column) {
            if (!device_column)
                return false;

            GPU_Typed_Base_Column<float>* d_col_ptr = static_cast<GPU_Typed_Base_Column<float>*> (device_column);
            if (!d_col_ptr)
                return false;

            return d_col_ptr->insert_from_device_memory_in_host_vector(host_column);
        }

    }; //end namespace gpu
}; //end namespace CogaDB


