
#include <thrust/copy.h>
#include <thrust/gather.h>
#include <gpu/gpu_positionlist.hpp>
#include <cstdlib>
#include <assert.h>

namespace CoGaDB{
	namespace gpu{	

			Impl_GPU_Positionlist::Impl_GPU_Positionlist(){
			
			}

			Impl_GPU_Positionlist::Impl_GPU_Positionlist(const std::vector<TID>* pos_list) {

				assert(pos_list!=NULL);

				pos_list_.resize(pos_list->size());
				thrust::copy(pos_list->begin(),pos_list->end(),pos_list_.begin());
	
			}

			Impl_GPU_Positionlist::Impl_GPU_Positionlist(const thrust::device_vector<TID>& pos_list) : pos_list_(pos_list){

			}

			Impl_GPU_Positionlist::~Impl_GPU_Positionlist(){

			}

			void Impl_GPU_Positionlist::clear(){
				pos_list_.clear();
			}
			
			const GPU_PositionlistPtr Impl_GPU_Positionlist::aggregate(GPU_PositionlistPtr dev_tids) const{ //dev_tids indexes the position list this points to, therefore dev_tids is the map and this->pos_list_ are the values of the gather operation, whereas the new created GPU position list is the out put
				shared_pointer_namespace::shared_ptr<Impl_GPU_Positionlist> dev_pos_list = shared_pointer_namespace::static_pointer_cast<Impl_GPU_Positionlist> (dev_tids);

				Impl_GPU_Positionlist* aggregated_pos_list = new Impl_GPU_Positionlist();

				aggregated_pos_list->pos_list_.resize(dev_pos_list->pos_list_.size());
				//std::cout << "Calling Gather..." << std::endl;
				 thrust::gather(dev_pos_list->pos_list_.begin(), dev_pos_list->pos_list_.end(), // begin and end of map
                				 this->pos_list_.begin(), //from this object
                 				 aggregated_pos_list->pos_list_.begin()); //from created column
            	//std::cout << "Result size: " << dev_pos_list->pos_list_.size() << " (" << dev_pos_list->pos_list_.end()-dev_pos_list->pos_list_.begin() << ")" << std::endl;     				 		std::cout << "Result size stored in vector!: " << aggregated_pos_list->pos_list_.size() << std::endl;  
				return GPU_PositionlistPtr(aggregated_pos_list);
			}

			bool Impl_GPU_Positionlist::insert(std::vector<TID>* pos_list){
				assert(pos_list!=NULL);

				pos_list_.resize(pos_list->size());
				thrust::copy(pos_list->begin(),pos_list->end(),pos_list_.begin());
	
				return true;
			}

			size_t Impl_GPU_Positionlist::size() const throw(){
				return pos_list_.size();
			}

			const std::type_info& Impl_GPU_Positionlist::type() const throw(){
				return typeid(Impl_GPU_Positionlist);
			}


			const boost::any Impl_GPU_Positionlist::get(TID tid){
				return boost::any(pos_list_[tid]);
			}


			const PositionListPtr Impl_GPU_Positionlist::toHostPositionlist() const{
				//std::cout << "toHostPositionlist: " << std::endl;
				PositionListPtr host_vec_ptr = PositionListPtr(new std::vector<TID>(pos_list_.size()));
				
				//std::cout << "host_vec_ptr size: " << host_vec_ptr << std::endl;
				thrust::copy(pos_list_.begin(),pos_list_.end(),host_vec_ptr->begin());
				return host_vec_ptr;
			}

			void Impl_GPU_Positionlist::print() const throw(){
				thrust::host_vector<TID> v(pos_list_);
				//std::cout << "GPU Position List: " << std::endl;
				for(unsigned int i=0;i<v.size();i++){
					std::cout << v[i] << std::endl;
				}
			}

			thrust::device_vector<TID>& Impl_GPU_Positionlist::getContent(){
				return pos_list_;
			}


	}; //end namespace gpu
}; //end namespace CogaDB

