#include <thrust/version.h>

//thrust includes
#include <thrust/device_vector.h>
#include <thrust/copy.h>
#include <thrust/sort.h>
#include <thrust/partition.h>
#include <thrust/functional.h>

#include <gpu/gpu_algorithms.hpp>
#include <gpu/gpu_typed_base_column.hpp>

using namespace std;

namespace CoGaDB{
	namespace gpu{	

const std::pair<GPU_Base_ColumnPtr,GPU_Base_ColumnPtr> aggregation(shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_key_col, shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_value_col, AggregationMethod agg_meth){


	thrust::device_vector<int>& keys = dev_key_col->getContent();
	thrust::device_vector<int>& values = dev_value_col->getContent();

	//groupby
	typedef thrust::tuple<thrust::device_vector<int>::iterator,thrust::device_vector<int>::iterator> Key_Value_Iterator_Tuple;

	thrust::zip_iterator<Key_Value_Iterator_Tuple> first = thrust::make_zip_iterator(thrust::make_tuple(keys.begin(), values.begin()));
	thrust::zip_iterator<Key_Value_Iterator_Tuple> last  = thrust::make_zip_iterator(thrust::make_tuple(keys.end(), values.end()));

	//expects data sorted after group_column (keys)
	//thrust::stable_sort(first,last);

	//here starts aggregation

	//std::assert(keys.size()==values.size());

	GPU_Typed_Base_Column<int>* result_key_column = new GPU_Typed_Base_Column<int>(dev_key_col->getName(), dev_key_col->getType());
	GPU_Typed_Base_Column<int>* result_value_column = new GPU_Typed_Base_Column<int>(dev_value_col->getName(), dev_value_col->getType());

//	thrust::device_vector<int> result_keys(keys.size());
//	thrust::device_vector<int> result_values(values.size());


	thrust::device_vector<int>& result_keys=result_key_column->getContent();
	thrust::device_vector<int>& result_values=result_value_column->getContent();

	result_keys.resize(keys.size());
	result_values.resize(values.size());

//  const int N = 7;
//  int A[N] = {1, 3, 3, 3, 2, 2, 1}; // input keys
//  int B[N] = {9, 8, 7, 6, 5, 4, 3}; // input values
//  int C[N];                         // output keys
//  int D[N];                         // output values

  thrust::pair<thrust::device_vector<int>::iterator,thrust::device_vector<int>::iterator> new_end;

  if(agg_meth==SUM){
  thrust::equal_to<int> binary_pred;
  thrust::plus<int> binary_op;
  new_end = thrust::reduce_by_key(keys.begin(), keys.end(), values.begin(), result_keys.begin(), result_values.begin() , binary_pred, binary_op);
  }else if(agg_meth==MIN){
    thrust::equal_to<int> binary_pred;
    thrust::minimum<int> binary_op;
    new_end = thrust::reduce_by_key(keys.begin(), keys.end(), values.begin(), result_keys.begin(), result_values.begin() , binary_pred, binary_op);
  }else if(agg_meth==MAX){
    thrust::equal_to<int> binary_pred;
    thrust::maximum<int> binary_op;
    new_end = thrust::reduce_by_key(keys.begin(), keys.end(), values.begin(), result_keys.begin(), result_values.begin() , binary_pred, binary_op);
  //}else if(agg_meth==){

  }else{
    std::cerr << "FATAL Error! Unknown Aggregation Method!" << agg_meth << endl;

  }

	result_keys.resize(new_end.first-result_keys.begin());
	result_values.resize(new_end.first-result_keys.begin());
	
  
//  for(int i=0;i<new_end.first-result_keys.begin();i++){
//  		std::cout << "result key: " << result_keys[i] << " value: " << result_values[i] << std::endl;
//  }
	assert(result_key_column!=NULL);
	assert(result_value_column!=NULL);
  return std::pair<GPU_Base_ColumnPtr,GPU_Base_ColumnPtr>(GPU_Base_ColumnPtr(result_key_column),GPU_Base_ColumnPtr(result_value_column));

}


const std::pair<GPU_Base_ColumnPtr,GPU_Base_ColumnPtr> GPU_Operators::groupby(GPU_Base_ColumnPtr grouping_column, 
																										 GPU_Base_ColumnPtr aggregation_column,  
																										 AggregationMethod agg_meth){

				if(!grouping_column || !aggregation_column)
					return std::pair<GPU_Base_ColumnPtr,GPU_Base_ColumnPtr>();

				//string name = device_column->getName();
				//AttributeType type = device_column->getType();
				//cout << "Processing Column " << name << " with type " << type << endl;

				std::pair<GPU_Base_ColumnPtr,GPU_Base_ColumnPtr> result;
				string name = aggregation_column->getName();
				AttributeType type = aggregation_column->getType();
				//cout << name << ":" << type << endl;
				if(type==INT){
					if(!quiet && verbose)
					cout << "Execute Groupby on GPU (int)" << endl;
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_key_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (grouping_column);
					shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<int> > dev_value_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<int> > (aggregation_column);
	
					result=aggregation(dev_key_col,dev_value_col,agg_meth);

				}else if(type==FLOAT){
					//shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_key_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (grouping_column);
					//shared_pointer_namespace::shared_ptr<GPU_Typed_Base_Column<float> > dev_value_col = shared_pointer_namespace::static_pointer_cast<GPU_Typed_Base_Column<float> > (aggregation_column);
	
					//result=aggregation(dev_key_col,dev_value_col,agg_meth);		
				}else if(type==VARCHAR){
					cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name 
			  			  << " Note: VARCHAR is currently not supported, will be added again in the future!"<< endl;
				}else if(type==BOOLEAN){
					//ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
					cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name 
						  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
				}else{
					cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << name << endl;
				}

				return result;

}

	}; //end namespace gpu
}; //end namespace CoGaDB

