
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gpu/gpu_algorithms.hpp>
#include <cuda.h>
//#include <cuda_runtime_api.h>

namespace CoGaDB{

#ifdef HAVE_CUDAMEMGETINFO

unsigned int getTotalGPUMemorySizeInByte(){

//get the amount of free and total memory of the device
size_t available_memory_in_bytes=0, total_memory_in_bytes=0;
if(cudaSuccess!=cudaMemGetInfo(&available_memory_in_bytes, &total_memory_in_bytes)){
	std::cout << "Error while trying to retrieve memory size information! Aborting..." << std::endl;
	std::exit(-1);
	return 0;
 }
 return total_memory_in_bytes;
}

unsigned int getFreeGPUMemorySizeInByte(){

 //get the amount of free and total memory of the device
size_t available_memory_in_bytes=0, total_memory_in_bytes=0;
if(cudaSuccess!=cudaMemGetInfo(&available_memory_in_bytes, &total_memory_in_bytes)){
	std::cout << "Error while trying to retrieve memory size information! Aborting..." << std::endl;
	std::exit(-1);
	return 0;
 }

 return available_memory_in_bytes;
}

#else /* !HAVE_CUDAMEMGETINFO */

unsigned int getTotalGPUMemorySizeInByte() {
	return 0;
}

unsigned int getFreeGPUMemorySizeInByte() {
	return 0;
}

#endif /* HAVE_CUDAMEMGETINFO */



void printGPUStatus(){
 std::cout << "Number of GPU Cores: " << getNumberofGPUMultiprocessors() <<  std::endl;
 std::cout << "Total GPU RAM: " << getTotalGPUMemorySizeInByte() <<  std::endl;
 std::cout << "Available GPU RAM: " << getFreeGPUMemorySizeInByte() <<  std::endl;
}

unsigned int getNumberofGPUMultiprocessors(){

 cudaDeviceProp prop;
 cudaGetDeviceProperties(&prop,0);

 return prop.multiProcessorCount;
}

}; //end namespace CogaDB

