
#include <core/global_definitions.hpp>
#include <hardware_optimizations/simd_acceleration.hpp>


namespace CoGaDB{


    
#ifdef ENABLE_SIMD_ACCELERATION 

#include <stdint.h>

//#define __SSE__
//#define __SSE2__
//#define __SSE3__
//#define __SSE4_1__
//#define __MXX__

// SSE compiler intrinsics
#ifdef __SSE__
#include <xmmintrin.h>
#endif

//For SSE2:
#ifdef __SSE2__
extern "C"
{
#include <emmintrin.h>
#include <mmintrin.h>
}
#endif

//For SSE3:
#ifdef __SSE3__
extern "C" 
{
#include <pmmintrin.h>
#include <immintrin.h>   // (Meta-header, for GCC only)
}
#endif

//For SSE4: (WITHOUT extern "C") 
#ifdef __SSE4_1__
#include <smmintrin.h>
#endif


//#define COGADB_SIMD_SCAN_INT(array, array_size, result_array, comparison_value,SIMD_COMPARISON_FUNCTION,COMPARISON_OPERATOR, result_size) unsigned int pos=0; \
//	__m128i* sse_array = reinterpret_cast<__m128i*>(array); \
//	assert(sse_array!=NULL); \
//        const int sse_array_length = (array_size*sizeof(int))/sizeof(__m128i); \
//        __m128i comp_val=_mm_set1_epi32(comparison_value); \
//        __m128i read_value=_mm_set1_epi32 (0); \
//        for(unsigned int i=0;i<sse_array_length;i++){ \
//		read_value=_mm_loadu_si128(&sse_array[i]); \
//                __m128 comp_result = (__m128) SIMD_COMPARISON_FUNCTION(read_value,comp_val); \
//                int mask= _mm_movemask_ps(comp_result); \
//                if(mask){ \
//                    for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ \
//                        if((mask >> j) & 1){ /*jth bit*/ \
//                            result_array[pos++]=i*(sizeof(__m128i)/sizeof(int))+j; \
//                        }  \
//                    }  \
//                }  \
//	}  \
//        for(unsigned int i=(((array_size*sizeof(int))/sizeof(__m128i))*sizeof(__m128i))/sizeof(int);i<array_size;i++){  \
//            if(array[i] COMPARISON_OPERATOR comparison_value){  \
//                result_array[pos++]=i;  \
//            }  \
//        }\
//        result_size=pos;

//#define COGADB_SIMD_SCAN_INT(array, array_size, result_array, comparison_value,SIMD_COMPARISON_FUNCTION,COMPARISON_OPERATOR, result_size) \
//        __m128i* sse_array = reinterpret_cast<__m128i*>(array); \
//	assert(sse_array!=NULL); \
//	int alignment_offset = ((int)sse_array)%sizeof(__m128i); \
//	const int sse_array_length = (array_size-alignment_offset)*sizeof(int)/sizeof(__m128i); \
//	char* tmp_array = (char*) sse_array; \
//	tmp_array+=alignment_offset; \
//	sse_array=reinterpret_cast<__m128i*>(tmp_array); \
//	unsigned int pos=0; \
//	__m128i comp_val=_mm_set1_epi32(comparison_value); \
//	__m128i read_value=_mm_set1_epi32 (0); \
//	if(alignment_offset!=0){ \
//            for(unsigned int i=0;i<alignment_offset;i++){ \
//                if(array[i] COMPARISON_OPERATOR comparison_value){ \
//                    result_array[pos++]=i; \
//                } \
//            } \
//        } \
//        for(unsigned int i=0;i<sse_array_length;i++){ \
//                assert(((int)sse_array)%sizeof(__m128i)==0); \
//                read_value=_mm_load_si128(&sse_array[i]); \
//                __m128 comp_result = (__m128) SIMD_COMPARISON_FUNCTION(read_value,comp_val); \
//                int mask= _mm_movemask_ps(comp_result); \
//                if(mask){ \
//                    for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ \
//                        if((mask >> j) & 1){ \
//                            result_array[pos++]=i*(sizeof(__m128i)/sizeof(int))+j+alignment_offset; \
//                        } \
//                    } \
//                } \
//	} \
//        for(unsigned int i=(sse_array_length*sizeof(__m128i)/sizeof(int))+alignment_offset;i<array_size;i++){ \
//            if(array[i] COMPARISON_OPERATOR comparison_value){ \
//                result_array[pos++]=i; \
//            } \
//        } \
//        result_size=pos;


#define SIMD_DEBUG_MODE false

#define COGADB_SIMD_SCAN_INT(array, array_size, result_array, comparison_value,SIMD_COMPARISON_FUNCTION,COMPARISON_OPERATOR, result_size) \
        __m128i* sse_array = reinterpret_cast<__m128i*>(array); \
	assert(sse_array!=NULL); \
	int alignment_offset = ((intptr_t)sse_array)%sizeof(__m128i); \
	const int sse_array_length = (array_size-alignment_offset)*sizeof(int)/sizeof(__m128i); \
        if(SIMD_DEBUG_MODE) std::cout << "SSE Array Length: " << sse_array_length << std::endl; \
	char* tmp_array = (char*) sse_array; \
	tmp_array+=alignment_offset; \
	sse_array=reinterpret_cast<__m128i*>(tmp_array); \
        if(SIMD_DEBUG_MODE)  std::cout << "array adress: "  << (void*)array <<  "sse array: " << (void*)sse_array << std::endl; \
        if(SIMD_DEBUG_MODE)  std::cout << "First SSE Array Element: " << ((int*)sse_array)[0] << std::endl; \
	unsigned int pos=0; \
	__m128i comp_val=_mm_set1_epi32(comparison_value); \
	__m128i read_value=_mm_set1_epi32 (0); \
        if(SIMD_DEBUG_MODE)  std::cout << "alignment_offset " << alignment_offset << std::endl; \
        if(alignment_offset!=0){ \
            if(SIMD_DEBUG_MODE) std::cout << "process first unaligned data chunk: index 0 to " << alignment_offset << std::endl; \
            for(unsigned int i=0;i<alignment_offset/sizeof(int);i++){ \
                if(SIMD_DEBUG_MODE){ \
                    std::cout << "index "<< i << std::endl; \
                    std::cout << "value " << array[i] << " match:" << (array[i] COMPARISON_OPERATOR comparison_value) << std::endl; \
                } \
                if(array[i] COMPARISON_OPERATOR comparison_value){ \
                    result_array[pos++]=i; \
                } \
            } \
        } \
        for(unsigned int i=0;i<sse_array_length;i++){ \
                assert(((intptr_t)sse_array)%sizeof(__m128i)==0); \
                read_value=_mm_load_si128(&sse_array[i]); \
                if(SIMD_DEBUG_MODE){ \
                    std::cout << "index: " << i << std::endl; \
                } \
                __m128 comp_result = (__m128) SIMD_COMPARISON_FUNCTION(read_value,comp_val); \
                int mask= _mm_movemask_ps(comp_result); \
                if(SIMD_DEBUG_MODE) std::cout << "Mask: " << std::hex << mask << std::dec << std::endl; \
                if(mask){ \
                    if(SIMD_DEBUG_MODE) std::cout << "at least one match!" << std::endl; \
                    for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ \
                        if(SIMD_DEBUG_MODE) std::cout << "sub index: " << j << " value: " << ((mask >> j) & 1) << std::endl; \
                        int tmp =((mask >> j) & 1); \
                        result_array[pos]=i*(sizeof(__m128i)/sizeof(int))+j+(alignment_offset/sizeof(int)); \
                        pos+=tmp; \
                    } \
                } \
	} \
        if(SIMD_DEBUG_MODE) std::cout << "Remaining offsets: " << (sse_array_length*sizeof(__m128i)/sizeof(int))+alignment_offset  << " to " << array_size << std::endl; \
        for(unsigned int i=(sse_array_length*sizeof(__m128i)/sizeof(int))+(alignment_offset/sizeof(int));i<array_size;i++){ \
            if(array[i] COMPARISON_OPERATOR comparison_value){ \
                result_array[pos++]=i; \
            } \
        } \
        result_size=pos;

#define COGADB_SIMD_SCAN_FLOAT(array, array_size, result_array, comparison_value,SIMD_COMPARISON_FUNCTION,COMPARISON_OPERATOR, result_size) \
   __m128* sse_array = reinterpret_cast<__m128*>(array); \
	assert(sse_array!=NULL); \
	int alignment_offset = ((intptr_t)sse_array)%sizeof(__m128); \
	const int sse_array_length = (array_size-alignment_offset)*sizeof(int)/sizeof(__m128); \
        if(SIMD_DEBUG_MODE) std::cout << "SSE Array Length: " << sse_array_length << std::endl; \
	char* tmp_array = (char*) sse_array; \
	tmp_array+=alignment_offset; \
	sse_array=reinterpret_cast<__m128*>(tmp_array); \
        if(SIMD_DEBUG_MODE)  std::cout << "array adress: "  << (void*)array <<  "sse array: " << (void*)sse_array << std::endl; \
        if(SIMD_DEBUG_MODE)  std::cout << "First SSE Array Element: " << ((int*)sse_array)[0] << std::endl; \
	unsigned int pos=0; \
	__m128 comp_val=_mm_set1_ps(comparison_value); \
	__m128 read_value=_mm_set1_ps (0); \
        if(SIMD_DEBUG_MODE)  std::cout << "alignment_offset " << alignment_offset << std::endl; \
        if(alignment_offset!=0){ \
            if(SIMD_DEBUG_MODE) std::cout << "process first unaligned data chunk: index 0 to " << alignment_offset << std::endl; \
            for(unsigned int i=0;i<alignment_offset/sizeof(int);i++){ \
                if(SIMD_DEBUG_MODE){ \
                    std::cout << "index "<< i << std::endl; \
                    std::cout << "value " << array[i] << " match:" << (array[i] COMPARISON_OPERATOR comparison_value) << std::endl; \
                } \
                if(array[i] COMPARISON_OPERATOR comparison_value){ \
                    result_array[pos++]=i; \
                } \
            } \
        } \
        for(unsigned int i=0;i<sse_array_length;i++){ \
                assert(((intptr_t)sse_array)%sizeof(__m128)==0); \
                read_value= _mm_load_ps((float*)&sse_array[i]); \
                if(SIMD_DEBUG_MODE){ \
                    std::cout << "index: " << i << std::endl; \
                } \
                __m128 comp_result = SIMD_COMPARISON_FUNCTION(read_value,comp_val); \
                int mask= _mm_movemask_ps(comp_result); \
                if(SIMD_DEBUG_MODE) std::cout << "Mask: " << std::hex << mask << std::dec << std::endl; \
                if(mask){ \
                    if(SIMD_DEBUG_MODE) std::cout << "at least one match!" << std::endl; \
                    for(unsigned j=0;j<sizeof(__m128)/sizeof(int);++j){ \
                        if(SIMD_DEBUG_MODE) std::cout << "sub index: " << j << " value: " << ((mask >> j) & 1) << std::endl; \
                        int tmp =((mask >> j) & 1); \
                        result_array[pos]=i*(sizeof(__m128)/sizeof(int))+j+(alignment_offset/sizeof(int)); \
                        pos+=tmp; \
                    } \
                } \
	} \
        if(SIMD_DEBUG_MODE) std::cout << "Remaining offsets: " << (sse_array_length*sizeof(__m128)/sizeof(int))+alignment_offset  << " to " << array_size << std::endl; \
        for(unsigned int i=(sse_array_length*sizeof(__m128)/sizeof(int))+(alignment_offset/sizeof(int));i<array_size;i++){ \
            if(array[i] COMPARISON_OPERATOR comparison_value){ \
                result_array[pos++]=i; \
            } \
        } \
        result_size=pos;




/*
//
//                if(mask){ \
//                    if(SIMD_DEBUG_MODE) std::cout << "at least one match!" << std::endl; \
//                    for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ \
//                        if(SIMD_DEBUG_MODE) std::cout << "sub index: " << j << " value: " << ((mask >> j) & 1) << std::endl; \
//                        if((mask >> j) & 1){ \
//                            result_array[pos++]=i*(sizeof(__m128i)/sizeof(int))+j+(alignment_offset/sizeof(int)); \
//                        } \
//                    } \
//                } */

#endif







#ifdef ENABLE_SIMD_ACCELERATION
        const PositionListPtr simd_selection_int(int* array, unsigned int array_size,  int comparison_value, const ValueComparator comp){
    
            PositionListPtr result_tids(new PositionList());
            //PositionListPtr result_tids(new PositionList(array_size));
            //result_tids->reserve(array_size);
//            //ensure that at least one element is in output buffer, so we can get a pointer to it
            //result_tids->push_back(0);
            //get the pointer to the output buffer
            //unsigned int* result_array = hype::util::begin_ptr(*result_tids); //new int[array_size];  
            unsigned int* result_array = new unsigned int[array_size];  
            assert(result_array!=NULL);
            
            unsigned int result_size=0;

            if(comp==EQUAL){
                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmpeq_epi32,==,result_size);
            }else if(comp==LESSER){
                //COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmplt_epi32,<,result_size);
                    __m128i* sse_array = reinterpret_cast<__m128i*>(array); 
                    assert(sse_array!=NULL); 
                    int alignment_offset = ((intptr_t)sse_array)%sizeof(__m128i);
                    const int sse_array_length = (array_size-alignment_offset)*sizeof(int)/sizeof(__m128i); 
                    if(SIMD_DEBUG_MODE) std::cout << "SSE Array Length: " << sse_array_length << std::endl; 
                    char* tmp_array = (char*) sse_array; 
                    tmp_array+=alignment_offset; 
                    sse_array=reinterpret_cast<__m128i*>(tmp_array); 
                    if(SIMD_DEBUG_MODE)  std::cout << "array adress: "  << (void*)array <<  "sse array: " << (void*)sse_array << std::endl; 
                    if(SIMD_DEBUG_MODE)  std::cout << "First SSE Array Element: " << ((int*)sse_array)[0] << std::endl; 
                    unsigned int pos=0; 
                    __m128i comp_val=_mm_set1_epi32(comparison_value); 
                    __m128i read_value=_mm_set1_epi32 (0); 
                    if(SIMD_DEBUG_MODE)  std::cout << "alignment_offset " << alignment_offset << std::endl; 
                    if(alignment_offset!=0){ 
                        if(SIMD_DEBUG_MODE) std::cout << "process first unaligned data chunk: index 0 to " << alignment_offset << std::endl; 
                        for(unsigned int i=0;i<alignment_offset/sizeof(int);i++){ 
                            if(SIMD_DEBUG_MODE){ 
                                std::cout << "index "<< i << std::endl; 
                                std::cout << "value " << array[i] << " match:" << (array[i] < comparison_value) << std::endl; 
                            } 
                            if(array[i] < comparison_value){ 
                                result_array[pos++]=i; 
                            } 
                        } 
                    } 
                    for(unsigned int i=0;i<sse_array_length;i++){ 
                            assert(((intptr_t)sse_array)%sizeof(__m128i)==0);
                            read_value=_mm_load_si128(&sse_array[i]); 
                            if(SIMD_DEBUG_MODE){ 
                                std::cout << "index: " << i << std::endl; 
                            } 
                            __m128 comp_result = (__m128) _mm_cmplt_epi32(read_value,comp_val); 
                            int mask= _mm_movemask_ps(comp_result); \
                            if(SIMD_DEBUG_MODE) std::cout << "Mask: " << std::hex << mask << std::dec << std::endl; 
                            if(mask){ 
                                if(SIMD_DEBUG_MODE) std::cout << "at least one match!" << std::endl; 
                                for(unsigned j=0;j<sizeof(__m128i)/sizeof(int);++j){ 
                                    if(SIMD_DEBUG_MODE) std::cout << "sub index: " << j << " value: " << ((mask >> j) & 1) << std::endl; 
                                    int tmp =((mask >> j) & 1); 
                                    result_array[pos]=i*(sizeof(__m128i)/sizeof(int))+j+(alignment_offset/sizeof(int)); 
                                    pos+=tmp; 
                                } 
                            } 
                    } 
                    if(SIMD_DEBUG_MODE) std::cout << "Remaining offsets: " << (sse_array_length*sizeof(__m128i)/sizeof(int))+alignment_offset  << " to " << array_size << std::endl; 
                    for(unsigned int i=(sse_array_length*sizeof(__m128i)/sizeof(int))+(alignment_offset/sizeof(int));i<array_size;i++){ 
                        if(array[i] < comparison_value){ 
                            result_array[pos++]=i; 
                        } 
                    } 
                    result_size=pos;
             }else if(comp==LESSER_EQUAL){
                //add one to comparison value to compare for less equal
                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value+1,_mm_cmplt_epi32,<,result_size);
            }else if(comp==GREATER){
                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value,_mm_cmpgt_epi32,>,result_size);
            }else if(comp==GREATER_EQUAL){
                //substract one of comparison value to compare for greater equal
                COGADB_SIMD_SCAN_INT(array,array_size,result_array,comparison_value-1,_mm_cmpgt_epi32,>,result_size);
            }else{
                COGADB_FATAL_ERROR("Invalid ValueComparator!","");
            }            
            
             
            
             //result_tids->resize(result_size);
             result_tids->insert(result_tids->end(),result_array,result_array+result_size);
             
             delete result_array;
             return result_tids;
        }
#endif           
        
#ifdef ENABLE_SIMD_ACCELERATION
//        template<>
//        inline const PositionListPtr Column<float>::selection(const boost::any& value_for_comparison, const ValueComparator comp){
//            //std::cout << "SIMD SCAN" << std::endl;
//            float value;
//            
//            if(value_for_comparison.type()!=typeid(float)){ 
//					 //allow comparison with itnegers as well
//                if(value_for_comparison.type()==typeid(int)){
//                    value = boost::any_cast<int>(value_for_comparison);
//                }else{
//                    COGADB_FATAL_ERROR(std::string("Typemismatch for column")+this->name_
//                                       +std::string(" Column Type: ") + typeid(float).name()
//                                       +std::string(" filter value type: ") + value_for_comparison.type().name(),""); 
//					 }
//            }else{
//                //everything fine, filter value matches type of column
//                value = boost::any_cast<float>(value_for_comparison);
//            }
//            unsigned int array_size = this->size();
//            float* array=hype::util::begin_ptr(values_);   
            
        const PositionListPtr simd_selection_float(float* array, unsigned int array_size,  float comparison_value, const ValueComparator comp){
                
            
            
            PositionListPtr result_tids(new PositionList());
            //result_tids->reserve(array_size);
//            //ensure that at least one element is in output buffer, so we can get a pointer to it
            //result_tids->push_back(0);
            //get the pointer to the output buffer
            unsigned int* result_array = new unsigned int[array_size];  //hype::util::begin_ptr(*result_tids); //new int[array_size];  
            assert(result_array!=NULL);
            //will be set by the makro
            unsigned int result_size=0;
            if(comp==EQUAL){
                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpeq_ps,==,result_size);
            }else if(comp==LESSER){
                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmplt_ps,<,result_size);
             }else if(comp==LESSER_EQUAL){
                //add one to comparison value to compare for less equal
                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmple_ps,<=,result_size);
            }else if(comp==GREATER){
                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpgt_ps,>,result_size);
            }else if(comp==GREATER_EQUAL){
                //substract one of comparison value to compare for greater equal
                COGADB_SIMD_SCAN_FLOAT(array,array_size,result_array,comparison_value,_mm_cmpge_ps,>=,result_size);
            }else{
                COGADB_FATAL_ERROR("Invalid ValueComparator!","");
            }            
            
            /**********************************************************************************/
            
 

            /**********************************************************************************/
            
             result_tids->insert(result_tids->end(),result_array,result_array+result_size);
             
             delete result_array;
            
             //result_tids->resize(result_size);
             return result_tids;
        }
#endif   

}; //end namespace CoGaDB
