
#include <lookup_table/lookup_column.hpp>
//#include <core/column_base_typed.hpp>
#include <iostream>

using namespace std;

namespace CoGaDB{

	LookupColumn::LookupColumn(TablePtr table, PositionListPtr tids) : table_(table), tids_(tids){

	}

	const ColumnPtr LookupColumn::getLookupArrayforColumnbyName(const Attribut& attr) const{
		string column_name = attr.second;
		//ColumnPtr col = table_->getColumnbyName(attr.second);

		const vector<ColumnPtr> columns = table_->getColumns();
		ColumnPtr col;
		for(unsigned int i=0;i<columns.size();i++){
			if(columns[i]->getName()==attr.second && columns[i]->getType()==attr.first){
				if(!quiet && verbose) cout << "Found Column " << attr.second << " in Table " << table_->getName() << endl;
				col=columns[i];
			}
		}

		if(!col){
			cout << "Error: Could not find Column " << attr.second << " in Table " << table_->getName() << endl;
			return ColumnPtr(); //std::tr1::shared_ptr<LookupArray<T> >(  ); //return NULL Pointer
		}
		if(col->getType()!=attr.first){
			cout << "FATAL ERROR: FOUND Column " << col->getName() << " of Table " << table_->getName() << " does not match Attribut type for LookupColumn!" << endl;
			cout << "File: " << __FILE__ << " Line:" << __LINE__ << endl;
			return ColumnPtr(); //return NULL Pointer
		}

//		if(typeid(T)!=col->type()){
//			return std::tr1::shared_ptr<LookupArray<T> >(  ); //return NULL Pointer
//		}
		AttributeType type=col->getType();

//	ColumnPtr ptr;
	if(type==INT){
		return ColumnPtr(new LookupArray<int>(column_name,INT,col,tids_));
	}else if(type==FLOAT){
		return ColumnPtr(new LookupArray<float>(column_name,FLOAT,col,tids_));
	}else if(type==VARCHAR){
		return ColumnPtr(new LookupArray<string>(column_name,VARCHAR,col,tids_));
	}else if(type==BOOLEAN){
		//return ColumnPtr(new LookupArray<bool>(column_name,BOOLEAN,col,tids_));
		//ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
		cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name 
			  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
	}else{
		cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name << endl;
	}


		return ColumnPtr();
	}

/*
	const ColumnVectorPtr LookupColumn::getDenseValueColumns() const{
		
		ColumnVectorPtr result(new ColumnVector());

		TableSchema schema = table_->getSchema();
		TableSchema::iterator it;
		for(it=schema.begin();it!=schema.end();it++){
			ColumnPtr col = getLookupArrayforColumnbyName(*it);
			if(!col){
				cout << "Error! LookupColumn requested column '" << it->second << "' which dos not exist in Table '" << table_->getName() << "'!" << endl;
				cout << "Schema inconsistent to original Table!" << endl;
				exit(-1);
			}
			if(!quiet && verbose) cout << "Create Lookup Array for Column " << it->second << " of table " << table_->getName() << endl;
			result->push_back(col);
		}
		return result;
	}*/

	const ColumnVectorPtr LookupColumn::getLookupArrays() const{
                //
            if(!quiet && verbose){
                cout << " LookupColumn::getLookupArrays(): ";
                if(table_->isMaterialized()){
                    cout << "Materialized Table: " << table_->getName() << endl;
                }else{
                    cout << "Lookup Table: " << table_->getName() << endl;
                }
            }
		ColumnVectorPtr result(new ColumnVector());

		TableSchema schema = table_->getSchema();
                //if(!quiet && verbose && debug) table_->printSchema();
		TableSchema::iterator it;
		for(it=schema.begin();it!=schema.end();it++){
                        if(!quiet && verbose && debug) cout << "Create LookupArray for Column " << table_->getName() << "." << it->second << endl;
			ColumnPtr col = getLookupArrayforColumnbyName(*it);
			if(!col){
				cout << "Error! LookupColumn requested column '" << it->second << "' which dos not exist in Table '" << table_->getName() << "'!" << endl;
				cout << "Schema inconsistent to original Table!" << endl;
				exit(-1);
			}
			if(!quiet && verbose) cout << "Create Lookup Array for Column " << it->second << " of table " << table_->getName() << endl;
			result->push_back(col);
		}
		return result;
	}

	/*
	const ColumnPtr LookupColumn::getLookupArrayforColumn(std::string column_name){
		
		ColumnPtr col = table_->getColumnbyName(column_name);

	AttributeType type=col->getType();

//	ColumnPtr ptr;
	if(type==INT){
		return ColumnPtr(new LookupArray<int>(column_name,INT,col,tids_));
	}else if(type==FLOAT){
		return ColumnPtr(new LookupArray<float>(column_name,FLOAT,col,tids_));
	}else if(type==VARCHAR){
		return ColumnPtr(new LookupArray<string>(column_name,VARCHAR,col,tids_));
	}else if(type==BOOLEAN){
		return ColumnPtr(new LookupArray<bool>(column_name,BOOLEAN,col,tids_));
		//ptr=ColumnPtr(new Column<bool>(name,BOOLEAN));
		cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name 
			  << " Note: bool is currently not supported, will be added again in the future!"<< endl;
	}else{
		cout << "Fatal Error! invalid AttributeType: " << type << " for Column: " << column_name << endl;
	}
		return ColumnPtr();
	}
	*/

	const TablePtr LookupColumn::getTable() const throw(){
		return table_;
	}

	const PositionListPtr LookupColumn::getPositionList() const throw(){
		return tids_;
	}
	//returns pointer to new LookupColumn which is the result of this call
	const shared_pointer_namespace::shared_ptr<LookupColumn> LookupColumn::aggregate(const LookupColumn& lookup_col) const{ //it would be sufficient if this was a Postitionlist
		const PositionListPtr lookup_col_tids = lookup_col.getPositionList();

		PositionListPtr aggregated_tids(new PositionList());
		const PositionListPtr current_tids = tids_;

		if(tids_->empty()){
			cout << "TID List is empty: " << endl;
			return shared_pointer_namespace::shared_ptr<LookupColumn>(new LookupColumn(table_, aggregated_tids) );			
		}

		TID tid(0), translated_tid(0);
		for(unsigned int i=0;i<lookup_col_tids->size();i++){
			tid=(*lookup_col_tids)[i];
			//translated_tid=(*current_tids)[tid];
			try{
			translated_tid=current_tids->at(tid);
			}catch(std::out_of_range& e){
				cout << e.what() << endl;
				cout << "BAD TID: " << tid << " in Lookup Column at position " << i << endl;
                     		cout << "In File " << __FILE__ << ":" << __LINE__ << endl;
				exit(-1);
			}
			aggregated_tids->push_back(translated_tid);
		}
		
		return shared_pointer_namespace::shared_ptr<LookupColumn>(new LookupColumn(table_, aggregated_tids) ); //table stays the same, hence pointer is passed to new lookup table, but positionlist is new, so pointer to new positionlist is passed
	}

	const LookupColumnPtr LookupColumn::copy() const{
		PositionListPtr new_tids (new PositionList(*tids_)); //call copy constructor of Positionlist (std::vector<TID>)
		return LookupColumnPtr( new LookupColumn(table_, new_tids) );
	}


}; //end namespace CogaDB

