#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <core/table.hpp>
#include <parser/commandline_interpreter.hpp>
#include <boost/tokenizer.hpp>


#include <unittests/unittests.hpp>
#include <util/tpch_benchmark.hpp>
#include <util/star_schema_benchmark.hpp>
#include <core/runtime_configuration.hpp>

#include <boost/lexical_cast.hpp>

#include <boost/random.hpp>
#include <boost/generator_iterator.hpp>
#include <boost/algorithm/string.hpp>

#include <utility>
#include <string>
#include <algorithm>

#include <parser/generated/Parser.h>
#include <query_processing/query_processor.hpp>
#include <core/gpu_column_cache.hpp>

#include <sql/server/sql_driver.hpp>

#ifdef LIBREADLINE_FOUND
#include <stdio.h>
#include <readline/readline.h>
#include <readline/history.h>
#endif

using namespace std;

namespace CoGaDB {
    using namespace query_processing;
    //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE);

    
//    bool optimize_and_execute(const std::string& query_name, LogicalQueryPlan& log_plan){
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            log_plan.print();
//        }
//        
//        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            cout << "Executing Query " << query_name << ":" << endl;
//            plan->print();
//        }
//        //execute the query
//        plan->run();
//        if(RuntimeConfiguration::instance().getPrintQueryPlan()){
//            plan->print();
//        }
//        TablePtr result = plan->getResult();
//        if(!result) return false;
//        result->print();
//        return true;
//    }
    
    
    bool printStatusOfGPUCache() {
        GPU_Column_Cache::instance().printStatus();
        return true;
    }

    bool simple_SSB_Query_Selection() {
        //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE); //hype::CPU_ONLY); //if you have only a CPU
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
                
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));


        //Attribut(INT,"YEARMONTHNUM")); //  numeric (YYYYMM) -- e.g. 199803

        boost::shared_ptr<logical_operator::Logical_Selection> selection_lineorder(new logical_operator::Logical_Selection("LO_DISCOUNT", boost::any(float(0)), GREATER, LOOKUP, default_device_constraint));

        selection_lineorder->setLeft(scan_lineorder);

        LogicalQueryPlan log_plan(selection_lineorder);
        log_plan.print();

        //cout << "Executing Query: " << endl;
        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();

        //Timestamp begin=getTimestamp();
        plan->run();
        //Timestamp end=getTimestamp();
        //assert(end>=begin);
        //cout << "Needed " << double(end-begin)/(1000*1000) << "ms ("<< double(end-begin)/(1000*1000*1000) <<"s) to process query '" << "1" <<  "' ..." << endl;
        //cout << "Result has " << plan->getResult()->getNumberofRows() << " rows" << endl;

        return true;
    }

    bool simple_SSB_Query_Join() {
        //hype::DeviceConstraint default_device_constraint(hype::ANY_DEVICE); //hype::CPU_ONLY); //if you have only a CPU
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_date(new logical_operator::Logical_Scan("DATE"));

        //Attribut(INT,"YEARMONTHNUM")); //  numeric (YYYYMM) -- e.g. 199803

        boost::shared_ptr<logical_operator::Logical_Selection> selection_date(new logical_operator::Logical_Selection("D_YEARMONTHNUM", boost::any(199803), EQUAL, LOOKUP, default_device_constraint));

        boost::shared_ptr<logical_operator::Logical_Join> join(new logical_operator::Logical_Join("D_DATEKEY", "LO_ORDERDATE", LOOKUP, hype::DeviceConstraint(hype::CPU_ONLY))); //GPU Join not supported


        selection_date->setLeft(scan_date);
        join->setLeft(selection_date);
        join->setRight(scan_lineorder);

        LogicalQueryPlan log_plan(join);
        log_plan.print();

        //cout << "Executing Query: " << endl;
        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();
        //Timestamp begin=getTimestamp();
        plan->run();
        //Timestamp end=getTimestamp();
        //assert(end>=begin);
        //cout << "Needed " << double(end-begin)/(1000*1000) << "ms ("<< double(end-begin)/(1000*1000*1000) <<"s) to process query '" << "1" <<  "' ..." << endl;
        //cout << "Result has " << plan->getResult()->getNumberofRows() << " rows" << endl;

        return true;
    }


    bool simple_SSB_Query_Aggregation() {
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        int value = 22;
        boost::shared_ptr<logical_operator::Logical_Scan> scan_lineorder(new logical_operator::Logical_Scan("LINEORDER"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_date(new logical_operator::Logical_Scan("DATE"));
        //boost::shared_ptr<logical_operator::Logical_ColumnConstantOperator>  column_constant_operation_sale(new logical_operator::Logical_ColumnConstantOperator("Price",value,"Add_Price_Value_LogOp",ADD,hype::DeviceConstraint(hype::CPU_ONLY)));

        boost::shared_ptr<logical_operator::Logical_Selection> selection_date(new logical_operator::Logical_Selection("D_YEAR", boost::any(1993), EQUAL, LOOKUP, default_device_constraint));
        boost::shared_ptr<logical_operator::Logical_Join> join(new logical_operator::Logical_Join("D_DATEKEY", "LO_ORDERDATE", LOOKUP, default_device_constraint)); //GPU Join not supported


        selection_date->setLeft(scan_date);
        join->setLeft(selection_date);
        join->setRight(scan_lineorder);

        std::list<std::string> column_list;
        column_list.push_back("D_YEAR");
        column_list.push_back("LO_EXTENDEDPRICE");
        column_list.push_back("LO_DISCOUNT");
        boost::shared_ptr<logical_operator::Logical_Projection> projection(new logical_operator::Logical_Projection(column_list));

        projection->setLeft(join);

        boost::shared_ptr<logical_operator::Logical_ColumnAlgebraOperator> column_algebra_operation_lineorder(new logical_operator::Logical_ColumnAlgebraOperator("LO_EXTENDEDPRICE", "LO_DISCOUNT", "REVENUE", MUL, default_device_constraint));

        column_algebra_operation_lineorder->setLeft(projection);


        std::list<std::string> sorting_column_names;
        sorting_column_names.push_back("D_YEAR");
        //column_names.push_back("Score");

        boost::shared_ptr<logical_operator::Logical_Sort> sort(new logical_operator::Logical_Sort(sorting_column_names, ASCENDING, LOOKUP, default_device_constraint)); //operation on String column only supported on CPU

        sort->setLeft(column_algebra_operation_lineorder);

        std::list<std::pair<string, AggregationMethod> > aggregation_functions;
        aggregation_functions.push_back(make_pair("REVENUE", SUM));

        boost::shared_ptr<logical_operator::Logical_Groupby> groupby(new logical_operator::Logical_Groupby(sorting_column_names, aggregation_functions, MATERIALIZE, default_device_constraint));

        groupby->setLeft(sort);

        LogicalQueryPlan log_plan(groupby); //column_algebra_operation_lineorder);
        log_plan.print();

        CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();
        plan->print();
        plan->run();

        if (!plan->getResult())
            return false;

        plan->getResult()->print();
        //boost::shared_ptr<logical_operator::Logical_Scan>  scan_sale(new logical_operator::Logical_Scan("Sale"));
        //boost::shared_ptr<logical_operator::Logical_ColumnAlgebraOperator>  column_algebra_operation_sale(new logical_operator::Logical_ColumnAlgebraOperator("Price","Sales","Add_Price_Sales_LogOp",ADD,hype::DeviceConstraint(hype::CPU_ONLY)));
        return true;
    }

    bool simple_SSB_Queries() {
        typedef bool (*SimpleCommandHandlerPtr)();
        typedef std::map<std::string, SimpleCommandHandlerPtr> SimpleCommandMap;
        SimpleCommandMap map;
//        map.insert(std::make_pair("Q1", &simple_SSB_Query_Join));
//        map.insert(std::make_pair("Q2", &simple_SSB_Query_Selection));
//        map.insert(std::make_pair("Q3", &simple_SSB_Query_Aggregation));
        map.insert(std::make_pair("ssb11", &SSB_Q11));
        map.insert(std::make_pair("ssb12", &SSB_Q12));
        map.insert(std::make_pair("ssb13", &SSB_Q13));
        map.insert(std::make_pair("ssb21", &SSB_Q21));
        map.insert(std::make_pair("ssb22", &SSB_Q22));
        map.insert(std::make_pair("ssb23", &SSB_Q23));
        map.insert(std::make_pair("ssb31", &SSB_Q31));
        map.insert(std::make_pair("ssb32", &SSB_Q32));
        map.insert(std::make_pair("ssb33", &SSB_Q33));
        map.insert(std::make_pair("ssb34", &SSB_Q34));
        map.insert(std::make_pair("ssb41", &SSB_Q41));
        map.insert(std::make_pair("ssb42", &SSB_Q42));
        map.insert(std::make_pair("ssb43", &SSB_Q43));        
        string input;
        cout << "Star Schema Query Test" << endl;
        cout << "Type 'back' to return to cmd" << endl;
        for (SimpleCommandMap::iterator it = map.begin(); it != map.end(); ++it) {
            cout << it->first << endl;
        }
        //workaroudn to use the CommandLineInterpreters History in this sub shell
        CommandLineInterpreter cmd;
        std::string prompt="CoGaDB>";
		
        while (true) {
            //cout << "CoGaDB>"; // << endl;
//            std::getline(std::cin, input);
//            if (cin.eof()) {
//                cout << endl;
//                return true;
//            }            
            if (cmd.getline(prompt, input)){ continue;}

            if (input == "back") return true;

            SimpleCommandMap::iterator it = map.find(input);
            if (it != map.end()) {
                cout << "Executing Query: " << it->first << endl;
                Timestamp begin = getTimestamp();
                if (!it->second()) cout << "Error: while processing Query '" << input << "'" << endl;
                Timestamp end = getTimestamp();
                cout << "Needed " << double(end - begin) / (1000 * 1000) << "ms (" << double(end - begin) / (1000 * 1000 * 1000) << "s) to process query '" << it->first << "' ..." << endl;

            } else {
                cout << "Query '" << input << "' not found" << endl;
            }
            //if(!cmd.execute(input)) cout << "Error! Command '" << input << "' failed!" << endl;

        }

        return true;
    }


    bool complex_tpch_queries() {
        hype::DeviceConstraint default_device_constraint = CoGaDB::RuntimeConfiguration::instance().getGlobalDeviceConstraint();
        cout << "Running Tests based on the TPC-H benchmark dataset" << endl;

        //scan operations
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_nation(new logical_operator::Logical_Scan("NATION"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_region(new logical_operator::Logical_Scan("REGION"));
        boost::shared_ptr<logical_operator::Logical_Scan> scan_part(new logical_operator::Logical_Scan("PART"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_supplier(new logical_operator::Logical_Scan("SUPPLIER"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_partsupp(new logical_operator::Logical_Scan("PARTSUPP"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_customer(new logical_operator::Logical_Scan("CUSTOMER"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_orders(new logical_operator::Logical_Scan("ORDERS"));
        //boost::shared_ptr<logical_operator::Logical_Scan> scan_lineitem(new logical_operator::Logical_Scan("LINEITEM"));


        /*
        Ich suche alle Lieferanten X, die ein Produkt von HERSTELLER Y verkaufen, welches einen Wert größer als Z hat. Dabei werden nur Produkte betrachtet, deren Lagerstand < A ist.
         */

        //T1 Suche alle Nationen einer Region -> EUROPE heraus
        //boost::shared_ptr<logical_operator::Logical_Selection> regionselection(new logical_operator::Logical_Selection("R_NAME", boost::any(std::string("EUROPE")), EQUAL, LOOKUP,  hype::DeviceConstraint(hype::CPU_ONLY)));
        //boost::shared_ptr<logical_operator::Logical_Join> regionnationjoin(new logical_operator::Logical_Join("R_REGIONKEY", "N_REGIONKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T2 Suche alle Produkte eines Herstellers heraus die einen gewissen Preis haben
        boost::shared_ptr<logical_operator::Logical_Selection> partselection1(new logical_operator::Logical_Selection("P_SIZE", boost::any(35), LESSER, LOOKUP, default_device_constraint));
        boost::shared_ptr<logical_operator::Logical_Selection> partselection2(new logical_operator::Logical_Selection("P_RETAILPRICE", boost::any(float(1100.99)), GREATER, LOOKUP, default_device_constraint));
        //T3 Suche Zuordnungen heraus, die eine bestimmte Anzahl haben und joine mit T2
        //boost::shared_ptr<logical_operator::Logical_Selection> partsuppselection(new logical_operator::Logical_Selection("PS_AVAILQTY", boost::any(1000), LESSER, LOOKUP, default_device_constraint));
        //boost::shared_ptr<logical_operator::Logical_Join> partpartsuppjoin(new logical_operator::Logical_Join("P_PARTKEY", "PS_PARTKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T4 Suche alle Supplier einer Region heruas (mit Hilfe von T1)
        //boost::shared_ptr<logical_operator::Logical_Join> t1supplierjoin(new logical_operator::Logical_Join("N_NATIONKEY", "S_NATIONKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T5 führe T3 und T4 zusammen
        // boost::shared_ptr<logical_operator::Logitidscal_Join> t3t4join(new logical_operator::Logical_Join("S_SUPPKEY", "PS_SUPPKEY", LOOKUP, default_device_constraint)); //GPU Join not supported
        //T6 sortiere die Ergebnisse nach verfügbarer Menge
        //boost::shared_ptr<logical_operator::Logical_Sort> availqtysort(new logical_operator::Logical_Sort("PS_AVAILQTY", ASCENDING, LOOKUP, default_device_constraint)); //operation on String column only supported on CPU
        //T7 Pjection
        //std::list<std::string> column_list;
        // column_list.push_back("S_NAME");
        // column_list.push_back("P_NAME");
        //column_list.push_back("S_ACCTBAL");
        //column_list.push_back("P_RETAILPRICE");
        //boost::shared_ptr<logical_operator::Logical_Projection> projection(new logical_operator::Logical_Projection(column_list)); //GPU Projection not supported



        //Führe T1 aus
        //regionselection->setLeft(scan_region);

        //regionnationjoin->setLeft(regionselection);
        //regionnationjoin->setRight(scan_nation);
        //Führe T2 aus
        partselection1->setLeft(scan_part);
        partselection2->setLeft(partselection1);
        //Führe T3 aus
        //partsuppselection->setLeft(scan_partsupp);

        //partpartsuppjoin->setLeft(partselection2);
        //partpartsuppjoin->setRight(partsuppselection);
        //Führe T4 aus
        //t1supplierjoin->setLeft(regionnationjoin);
        //t1supplierjoin->setRight(scan_supplier);
        //Führe T5 aus
        //t3t4join->setLeft(t1supplierjoin);
        //t3t4join->setRight(partpartsuppjoin);
        //Führe T6 aus
        //availqtysort->setLeft(partpartsuppjoin);
        //Führe T7 aus
        //projection->setLeft(partselection2);


        //Erstelle logischen Plan
        LogicalQueryPlan log_plan(partselection2);

        log_plan.print();
        //Führe Physischen Plan 30 mal aus
        for (unsigned int i = 0; i < 30; i++) {
            cout << "Executing Query: " << endl;
            CoGaDB::query_processing::PhysicalQueryPlanPtr plan = log_plan.convertToPhysicalQueryPlan();

            plan->print();

            plan->run();

            //typename PhysicalQueryPlan::Type
            TablePtr ptr = plan->getResult();
            if (ptr) {
                // ptr->print();
            } else {
                cout << "Error while processing query: pointer to result table is NULL, no results could be retrieved" << endl;

                return false;
            }
        }

        return true;
    }

    bool printDatabaseSchema() {

        std::vector<TablePtr>& tables = getGlobalTableList();
        cout << "Tables in database '" << RuntimeConfiguration::instance().getPathToDatabase() << "':" << endl;

        for (unsigned int i = 0; i < tables.size(); i++) {

            cout << tables[i]->getName() << " (";

            TableSchema schema = tables[i]->getSchema();
            TableSchema::iterator it;
            for (it = schema.begin(); it != schema.end(); ++it) {

                if (it != schema.begin()) cout << ",";
                cout << it->second;

            }

            cout << ") [" << tables[i]->getNumberofRows() << " rows]" << endl;
        }
        return true;
    }

    bool toggleQueryChopping() {
        RuntimeConfiguration::instance().setQueryChoppingEnabled(!RuntimeConfiguration::instance().isQueryChoppingEnabled());
        if (RuntimeConfiguration::instance().isQueryChoppingEnabled()) {
            cout << "Query Chopping is now activated." << endl;
        } else {
            cout << "Query Chopping is now deactivated." << endl;
        }
		return true;
    }

    bool help() {

        cout << "supported commands:" << endl
                << "loaddatabase\tloads complete database in main memory" << endl
                << "unittests\tperforms a self check of CoGaDB" << endl
                << "printschema\tprints the schema of the active database" << endl
                << "showgpucache\tprints status information of the GPU column cache"
                << "simple_ssb_queries\tsimple demonstrator for queries on SSB Benchmark data set" << endl
                << "set <variablename>=<variablevalue>" << endl
                << "print <variable>\tprint value of variable" << endl
                << "\t" << "Variables:" << endl
                << "\t" << "path_to_database\tabsolute or relative path to directory where the database is stored" << endl
                << "create_tpch_database <path to *.tbl files>\timport tables of TPC-H benchmark in CoGaDB" << endl
                << "create_ssb_database <path to *.tbl files>\timport tables of star schema benchmark in CoGaDB" << endl
                << "exec <SQL>\tExecute SQL statements" << endl
                << "toggleQC\tToggle the state of Query Chopping activation. Per default QC is off."
                << "ssbXY \tExecute SSB-Query X.Y (X has to be a number between 1 and 4; Y has to be a number between 1 and 3 except when X is 3, in this case 4 is valid for Y as well)" << endl
                << "setdevice <DEVICE> \tSets the default device, which is used for execution. Possible values are 'cpu', 'gpu' or 'any' to use either the CPU or the GPU or both." << endl
                << "setparallelizationmode <PARALLELIZATION MODE> \tSets the default parallelization mode for Subplans generated during Two Phase Physical Optimization (TOPPO) in the second phase (currently only for complex selections). Valid values are 'serial' and 'parallel'"
                << "quit" << endl;

        return true;
    }

    bool setVariable(const std::string& input) {

        unsigned int tokencounter = 0;
        typedef boost::tokenizer<boost::char_separator<char> >
                tokenizer;
        boost::char_separator<char> sep("=");
        tokenizer tok(input, sep);
        std::string command;
        std::string parameter;
        for (tokenizer::iterator beg = tok.begin(); beg != tok.end(); ++beg) {
            //cout << *beg << "\n";
            if (tokencounter == 0) command = *beg;
            if (tokencounter == 1) parameter = *beg;
            tokencounter++;
        }

        if (tokencounter > 2) {
            cout << "Error! For Command SET: To many parameters: <variable>=<value>" << endl;
            return false;
        }

        cout << "SET Variable: '" << command << "' to '" << parameter << "'" << endl;

        if (command == "path_to_database") {

            RuntimeConfiguration::instance().setPathToDatabase(parameter);
        }

        return true;
    }

    bool setDefaultDevice(const std::string& input) {
        if (input == "cpu") {
            //default_device_constraint = hype::DeviceConstraint(hype::CPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::CPU_ONLY));
            cout << "Set device to CPU." << endl;
        } else if (input == "gpu") {
            //default_device_constraint = hype::DeviceConstraint(hype::GPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::GPU_ONLY));
            cout << "Set device to GPU." << endl;
        } else if (input == "any") {
            //default_device_constraint = hype::DeviceConstraint(hype::ANY_DEVICE);
            CoGaDB::RuntimeConfiguration::instance().setGlobalDeviceConstraint(hype::DeviceConstraint(hype::ANY_DEVICE));            
            cout << "Set device to any device." << endl;
        } else {
            cout << "No device with the given devicecode could be found. Try 'cpu', 'gpu' or 'any' to switch the defaultdevice." << endl;

            return false;
        }
        return true;
    }
    
    
        bool setDefaultParallelizationModeTwoPhasePhysicalOptimization(const std::string& input) {
        if (input == "serial") {
            //default_device_constraint = hype::DeviceConstraint(hype::CPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setParallelizationModeForTwoPhasePhysicalOptimizationQueryPlans(CoGaDB::SERIAL);
            cout << "Set TOPPO parallelization mode to serial." << endl;
        } else if (input == "parallel") {
            //default_device_constraint = hype::DeviceConstraint(hype::GPU_ONLY);
            CoGaDB::RuntimeConfiguration::instance().setParallelizationModeForTwoPhasePhysicalOptimizationQueryPlans(CoGaDB::PARALLEL);
            cout << "Set TOPPO parallelization mode to parallel." << endl;
        } else {
            cout << "Unknown TOPPO Parallelization mode! Try 'serial' or 'parallel' to switch the TOPPO (Two Phase Physical Optimization) parallelization mode." << endl;

            return false;
        }
        return true;
    }
    
    
    
    

    bool printVariableValue(const std::string& input) {

        if (input == "path_to_database") {

            cout << "'" << RuntimeConfiguration::instance().getPathToDatabase() << "'" << endl;
        }

        return true;
    }

    CommandLineInterpreter::CommandLineInterpreter() : simple_command_map_(), command_map_(), prompt_("CoGaDB>") {

        simple_command_map_.insert(std::make_pair("loaddatabase", &loadTables));
        simple_command_map_.insert(std::make_pair("unittests", &CoGaDB::unit_tests::executeUnitTests));
        simple_command_map_.insert(std::make_pair("printschema", &printDatabaseSchema));
        simple_command_map_.insert(std::make_pair("help", &help));
        simple_command_map_.insert(std::make_pair("simple_ssb_queries", &simple_SSB_Queries));
        simple_command_map_.insert(std::make_pair("showgpucache", &printStatusOfGPUCache));
        simple_command_map_.insert(std::make_pair("toggleQC", &toggleQueryChopping));
//        simple_command_map_.insert(std::make_pair("ssb11", &SSB_Q11));
//        simple_command_map_.insert(std::make_pair("ssb12", &SSB_Q12));
//        simple_command_map_.insert(std::make_pair("ssb13", &SSB_Q13));
//        simple_command_map_.insert(std::make_pair("ssb21", &SSB_Q21));
//        simple_command_map_.insert(std::make_pair("ssb22", &SSB_Q22));
//        simple_command_map_.insert(std::make_pair("ssb23", &SSB_Q23));
//        simple_command_map_.insert(std::make_pair("ssb31", &SSB_Q31));
//        simple_command_map_.insert(std::make_pair("ssb32", &SSB_Q32));
//        simple_command_map_.insert(std::make_pair("ssb33", &SSB_Q33));
//        simple_command_map_.insert(std::make_pair("ssb34", &SSB_Q34));
//        simple_command_map_.insert(std::make_pair("ssb41", &SSB_Q41));
//        simple_command_map_.insert(std::make_pair("ssb42", &SSB_Q42));
//        simple_command_map_.insert(std::make_pair("ssb43", &SSB_Q43));
        
        //workaround so ssb queries work in batch mode
        simple_command_map_.insert(std::make_pair("ssb11", &SSB_Q11));
        simple_command_map_.insert(std::make_pair("ssb12", &SSB_Q12));
        simple_command_map_.insert(std::make_pair("ssb13", &SSB_Q13));
        simple_command_map_.insert(std::make_pair("ssb21", &SSB_Q21));
        simple_command_map_.insert(std::make_pair("ssb22", &SSB_Q22));
        simple_command_map_.insert(std::make_pair("ssb23", &SSB_Q23));
        simple_command_map_.insert(std::make_pair("ssb31", &SSB_Q31));
        simple_command_map_.insert(std::make_pair("ssb32", &SSB_Q32));
        simple_command_map_.insert(std::make_pair("ssb33", &SSB_Q33));
        simple_command_map_.insert(std::make_pair("ssb34", &SSB_Q34));
        simple_command_map_.insert(std::make_pair("ssb41", &SSB_Q41));
        simple_command_map_.insert(std::make_pair("ssb42", &SSB_Q42));
        simple_command_map_.insert(std::make_pair("ssb43", &SSB_Q43));

        command_map_.insert(std::make_pair("set", &setVariable));
        command_map_.insert(std::make_pair("create_tpch_database", &Unittest_Create_TPCH_Database));
        command_map_.insert(std::make_pair("create_ssb_database", &Unittest_Create_Star_Schema_Benchmark_Database));
        command_map_.insert(std::make_pair("print", &printVariableValue));
        command_map_.insert(std::make_pair("setdevice", &setDefaultDevice));
        command_map_.insert(std::make_pair("setparallelizationmode",&setDefaultParallelizationModeTwoPhasePhysicalOptimization));

        command_map_.insert(std::make_pair("exec", &SQL::commandlineExec));


        //
        //	create_ssb_database /media/1A2AA9012AA8DAD1_/benchmarks/star_schema_benchmark/SF1/

        //	Unittest_Create_Star_Schema_Benchmark_Database("/media/1A2AA9012AA8DAD1_/benchmarks/star_schema_benchmark/SF1/");

        //Unittest_Create_TPCH_Database
    }

    bool CommandLineInterpreter::execute(const string& input) {
        string input_trimmed = boost::algorithm::trim_copy(input);

        if (input_trimmed.empty())
            return true;
        if (input_trimmed[0] == '#')
            return true;

        {
            SimpleCommandMap::iterator it = simple_command_map_.find(input_trimmed);

            if (it != simple_command_map_.end())
                return it->second();
        }

        size_t command_len = input_trimmed.find_first_of(" \t");
        size_t param_begin = input_trimmed.find_first_not_of(" \t", command_len + 1);

        string command = input.substr(0, command_len);
        string parameter = input.substr(param_begin);

        ParameterizedCommandMap::iterator it = command_map_.find(command);
        if (it != command_map_.end())
            return it->second(parameter);

        cout << "Error! Command '" << command << "' not found!" << endl;

        return false;
    }

#ifdef LIBREADLINE_FOUND

    bool CommandLineInterpreter::getline(const string &prompt, string &result) {
        static char *line = NULL;

        if (line)
            free(line);

        line = readline(prompt.c_str());
        if (!line)
            return true; /* EOF */

        if (*line)
            add_history(line);
        result = line;

        return false;
    }
#else /* !LIBREADLINE_FOUND */

    bool CommandLineInterpreter::getline(const string &prompt, string &result) {
        cout << prompt;
        std::getline(cin, result);

        return cin.eof();
    }
#endif /* !LIBREADLINE_FOUND */
} //end namespace CogaDB


