
#include <query_processing/column_processing/cpu_column_constant_filter_operator.hpp>
#include <util/hardware_detector.hpp>
#include <core/lookup_array.hpp>

#include "core/gpu_column_cache.hpp"

namespace CoGaDB {

    namespace query_processing {
        //Map_Init_Function init_function_column_constant_filter_operator=physical_operator::map_init_function_column_constant_filter_operator;

        namespace physical_operator {

            column_processing::cpu::TypedOperatorPtr create_CPU_column_constant_filter_operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child) {
                logical_operator::Logical_Column_Constant_Filter& log_sort_ref = static_cast<logical_operator::Logical_Column_Constant_Filter&> (logical_node);
                //std::cout << "create SCAN Operator!" << std::endl;
                
                if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                        exit(-1);
                }
                assert(right_child==NULL); //unary operator
                 
                return column_processing::cpu::TypedOperatorPtr(new CPU_column_constant_filter_operator(sched_dec, left_child, log_sort_ref.getPredicate()));
            }
            
            column_processing::cpu::TypedOperatorPtr create_GPU_column_constant_filter_operator(column_processing::cpu::TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, column_processing::cpu::TypedOperatorPtr left_child, column_processing::cpu::TypedOperatorPtr right_child) {
                logical_operator::Logical_Column_Constant_Filter& log_sort_ref = static_cast<logical_operator::Logical_Column_Constant_Filter&> (logical_node);
                //std::cout << "create SCAN Operator!" << std::endl;
                
                if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
                        exit(-1);
                }
                assert(right_child==NULL); //unary operator
                 
                return column_processing::cpu::TypedOperatorPtr(new GPU_column_constant_filter_operator(sched_dec, left_child, log_sort_ref.getPredicate()));
            }

            column_processing::cpu::Physical_Operator_Map_Ptr map_init_function_column_constant_filter_operator() {
                //std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
                column_processing::cpu::Physical_Operator_Map map;
                if (!quiet) std::cout << "calling map init function! (COLUMN_CONSTANT_FILTER OPERATION)" << std::endl;
                //hype::Scheduler::instance().addAlgorithm("SCAN","TABLE_SCAN",hype::CPU,"Least Squares 1D","Periodic Recomputation");
                hype::AlgorithmSpecification selection_alg_spec_cpu("CPU_COLUMN_CONSTANT_FILTER",
                        "COLUMN_CONSTANT_FILTER",
                        hype::Least_Squares_1D,
                        hype::Periodic);

                hype::AlgorithmSpecification selection_alg_spec_gpu("GPU_COLUMN_CONSTANT_FILTER",
                        "COLUMN_CONSTANT_FILTER",
                        hype::Least_Squares_1D,
                        hype::Periodic);
                
                
                const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();

                for (unsigned int i = 0; i < dev_specs.size(); ++i) {
                    if (dev_specs[i].getDeviceType() == hype::CPU) {
                        hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu, dev_specs[i]);
                    }else if(dev_specs[i].getDeviceType() == hype::GPU){
                        hype::Scheduler::instance().addAlgorithm(selection_alg_spec_gpu, dev_specs[i]);
                    }
                }
                map["CPU_COLUMN_CONSTANT_FILTER"] = create_CPU_column_constant_filter_operator;
                map["GPU_COLUMN_CONSTANT_FILTER"] = create_GPU_column_constant_filter_operator;
                return column_processing::cpu::Physical_Operator_Map_Ptr(new column_processing::cpu::Physical_Operator_Map(map));
            }

            CPU_column_constant_filter_operator::CPU_column_constant_filter_operator(const hype::SchedulingDecision& sched_dec, TypedOperatorPtr child,
                    const Predicate& pred) : UnaryOperator<ColumnPtr, ColumnPtr>(sched_dec, child), pred_(pred) {
                //this->result_=getTablebyName(table_name_);
            }

            bool CPU_column_constant_filter_operator::execute() {
                if (!quiet && verbose && debug) std::cout << "Execute Filter Operation" << std::endl;

                assert(pred_.getPredicateType()==ValueConstantPredicate);
                ColumnPtr col = this->getInputData();
                //std::cout << "Input Column: " << col.get() << std::endl;
                //if(col==NULL) COGADB_FATAL_ERROR("Column_SCAN: Invalid Input Column! NULL Pointer...","");
                PositionListPtr tids = col->selection(pred_.getConstant(),pred_.getValueComparator());
                //create a LookupArray
                this->result_=createLookupArrayForColumn(col,tids);
                
//                   PredicateType getPredicateType() const;
//        const std::string& getColumn1Name() const;
//        const std::string& getColumn2Name() const;
//        const boost::any& getConstant() const;
//        ValueComparator getValueComparator() const;
//        void print() const;
//        std::string toString() const;        
                
                
                if (this->result_) {
                    return true;
                } else {
                    return false;
                }
            }

            CPU_column_constant_filter_operator::~CPU_column_constant_filter_operator(){
            }

            /*GPU Operation*/
            GPU_column_constant_filter_operator::GPU_column_constant_filter_operator(const hype::SchedulingDecision& sched_dec, TypedOperatorPtr child,
                    const Predicate& pred) : UnaryOperator<ColumnPtr, ColumnPtr>(sched_dec, child), pred_(pred) {
                //this->result_=getTablebyName(table_name_);
            }

            bool GPU_column_constant_filter_operator::execute() {
                if (!quiet && verbose && debug) std::cout << "Execute Filter Operation" << std::endl;

                assert(pred_.getPredicateType()==ValueConstantPredicate);
                //this->result_=this->getInputData()->selection(pred_.getConstant(),pred_.getValueComparator());
                ColumnPtr col = this->getInputData();
                //ensure that col is a dense value array, other columsn cannot be processed on GPU
                if(!col->isMaterialized()) col=col->materialize();
                //gpu::copy_column_host_to_device(ColumnPtr host_column)
                gpu::GPU_Base_ColumnPtr gpu_col = CoGaDB::GPU_Column_Cache::instance().getGPUColumn(col);
                if(!gpu_col){
                    std::cout << "Error! In GPU_column_constant_filter_operator::execute: Data Transfer to GPU Failed!" << std::endl;
                    return false;
                }
                gpu::GPU_PositionlistPtr gpu_tids = gpu::GPU_Operators::selection(gpu_col,pred_.getConstant(), pred_.getValueComparator());
                if(!gpu_col){
                    std::cout << "Error! In GPU_column_constant_filter_operator::execute: GPU Selection Operator Failed!" << std::endl;
                    return false;
                }                
                PositionListPtr tids = gpu::copy_PositionList_device_to_host(gpu_tids);
                if(!tids){
                    std::cout << "Error! In GPU_column_constant_filter_operator::execute: Data Transfer to CPU Failed!" << std::endl;
                    return false;
                }                            
                //create a LookupArray
                this->result_=createLookupArrayForColumn(col,tids);
     
                if (this->result_) {
                    return true;
                } else {
                    return false;
                }
            }

            GPU_column_constant_filter_operator::~GPU_column_constant_filter_operator(){
            }
            
        }//end namespace physical_operator


        namespace logical_operator {

            Logical_Column_Constant_Filter::Logical_Column_Constant_Filter(const Predicate& pred, hype::DeviceConstraint dev_constr) : TypedNode_Impl<ColumnPtr, physical_operator::map_init_function_column_constant_filter_operator>(false,dev_constr), pred_(pred){
            }

            unsigned int Logical_Column_Constant_Filter::getOutputResultSize() const {
                return 10;
            }

            double Logical_Column_Constant_Filter::getCalculatedSelectivity() const {
                return 0.3;
            }

            std::string Logical_Column_Constant_Filter::getOperationName() const {
                return "COLUMN_CONSTANT_FILTER";
            }

            const Predicate& Logical_Column_Constant_Filter::getPredicate() const{
                return pred_;
            }
            
            std::string Logical_Column_Constant_Filter::toString(bool verbose) const{
                std::string result="COLUMN_CONSTANT_FILTER";
                if(verbose){
                        result+=" (";
                        result+= pred_.toString();
                        result+=")";
                }
                return result;
            }

        }//end namespace logical_operator

    }//end namespace query_processing

}; //end namespace CogaDB
