
#include <query_processing/complex_selection_operator.hpp>
#include <util/hardware_detector.hpp>
#include <query_processing/query_processor.hpp>

namespace CoGaDB
{

namespace query_processing{

//Map_Init_Function init_function_ComplexSelection_operator=physical_operator::map_init_function_ComplexSelection_operator;

	namespace physical_operator{

                bool CPU_ComplexSelection_Operator::execute()
                {                    
#ifdef ENABLE_TWO_PHASE_PHYSICAL_OPTIMIZATION
                    //if there is one disjunction, we need the full query plan
                    if(knf_expr_.disjunctions.size()>1){
                        this->result_ = query_processing::two_phase_physical_optimization_selection(this->getInputData(), knf_expr_, dev_constr_, mat_stat_, CoGaDB::RuntimeConfiguration::instance().getParallelizationModeForTwoPhasePhysicalOptimizationQueryPlans());
                    //if we have just one disjunction, it is possible that we split up the query in selectiosn for single disjunctions, which we execute serially
                    //the other case is that the user specified a condition that can be represented by a single disjunction
                    //in both cases, it is correct to create a column based plan for the disjunction and construct a Lookup Table from the result    
                    }else if(knf_expr_.disjunctions.size()==1){
                        this->result_= BaseTable::selection(this->getInputData(), knf_expr_.disjunctions[0], mat_stat_);
                    }else{
                        COGADB_FATAL_ERROR("Complex Selection has empty conditions!","");
                    }
#else
                    this->result_ = BaseTable::selection(this->getInputData(), knf_expr_, mat_stat_, SERIAL);
#endif
                if (this->result_) {
                    setResultSize(((TablePtr) this->result_)->getNumberofRows());
                    return true;

                } else
                        return false;
                }    
            
	TypedOperatorPtr create_CPU_ComplexSelection_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_ComplexSelection& log_selection_ref = static_cast<logical_operator::Logical_ComplexSelection&>(logical_node);
		//std::cout << "create CPU_ComplexSelection_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
                
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new CPU_ComplexSelection_Operator(sched_dec,
		                        left_child,
		                        log_selection_ref.getKNF_Selection_Expression(),
                                        log_selection_ref.getDeviceConstraint(),
		                        log_selection_ref.getMaterializationStatus()) );
	}

	Physical_Operator_Map_Ptr map_init_function_complex_selection_operator()
	{
		//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
		Physical_Operator_Map map;
		if(!quiet) std::cout << "calling map init function for SELECTION operator!" << std::endl;
		//hype::Scheduler::instance().addAlgorithm("SELECTION","CPU_ComplexSelection_Algorithm",hype::CPU,"Least Squares 1D","Periodic Recomputation");
		//hype::Scheduler::instance().addAlgorithm("SELECTION","GPU_ComplexSelection_Algorithm",hype::GPU,"Least Squares 1D","Periodic Recomputation");

		hype::AlgorithmSpecification selection_alg_spec_cpu("CPU_ComplexSelection_Algorithm",
                                                                    "COMPLEX_SELECTION",
                                                                    hype::Least_Squares_1D,hype::Periodic);
                
		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();
		
		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu,dev_specs[i]);
                                //hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu_parallel,dev_specs[i]);
			}else if(dev_specs[i].getDeviceType()==hype::GPU){
				//hype::Scheduler::instance().addAlgorithm(selection_alg_spec_gpu,dev_specs[i]);			
			}
		}


		map["CPU_ComplexSelection_Algorithm"]=create_CPU_ComplexSelection_Operator;
                //map["CPU_ParallelComplexSelection_Algorithm"]=create_CPU_ParallelComplexSelection_Operator;
		//map["GPU_ComplexSelection_Algorithm"]=create_GPU_ComplexSelection_Operator;
		return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
	}

	}//end namespace physical_operator

}//end namespace query_processing

}; //end namespace CogaDB
