
#include <query_processing/groupby_operator.hpp>
#include <util/hardware_detector.hpp>

namespace CoGaDB
{

namespace query_processing{

//Map_Init_Function init_function_Groupby_operator=physical_operator::map_init_function_groupby_operator;

	namespace physical_operator{

	TypedOperatorPtr create_CPU_Groupby_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_Groupby& log_Groupby_ref = static_cast<logical_operator::Logical_Groupby&>(logical_node);
		if(!quiet && verbose)
			std::cout << "create CPU_Groupby_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new CPU_Groupby_Operator(sched_dec,
		                        left_child,
		                        log_Groupby_ref.getGroupingColumns(),
		                        log_Groupby_ref.getColumnAggregationFunctions(),
		                        log_Groupby_ref.getMaterializationStatus()) );
	}

	TypedOperatorPtr create_GPU_Groupby_Operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr left_child, TypedOperatorPtr right_child)
	{
		logical_operator::Logical_Groupby& log_Groupby_ref = static_cast<logical_operator::Logical_Groupby&>(logical_node);
		if(!quiet && verbose)
			std::cout << "create GPU_Groupby_Operator!" << std::endl;
		if(!left_child) {
                    std::cout << "Error! File: " << __FILE__ << " Line: " << __LINE__ << std::endl;
			exit(-1);
		}
		assert(right_child==NULL); //unary operator
		return TypedOperatorPtr(new GPU_Groupby_Operator(sched_dec,
		                        left_child,
		                        log_Groupby_ref.getGroupingColumns(),
		                        log_Groupby_ref.getColumnAggregationFunctions(),
		                        log_Groupby_ref.getMaterializationStatus()) );
	}



	Physical_Operator_Map_Ptr map_init_function_groupby_operator()
	{
		//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
		Physical_Operator_Map map;
		if(!quiet) std::cout << "calling map init function for GROUPBY operator!" << std::endl;
		//hype::Scheduler::instance().addAlgorithm("GROUPBY","CPU_Groupby_Algorithm",hype::CPU,"Least Squares 1D","Periodic Recomputation");
		//hype::Scheduler::instance().addAlgorithm("GROUPBY","GPU_Groupby_Algorithm",hype::GPU,"Least Squares 1D","Periodic Recomputation");


		
		hype::AlgorithmSpecification group_by_alg_spec_cpu("CPU_Groupby_Algorithm",
																	"GROUPBY",
																	hype::Least_Squares_1D,
																	hype::Periodic);	
	
		hype::AlgorithmSpecification group_by_alg_spec_gpu("GPU_Groupby_Algorithm",
																	"GROUPBY",
																	hype::Least_Squares_1D,
																	hype::Periodic);																		
																	

		//addAlgorithmSpecificationToHardware();
		
		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();
		
		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(group_by_alg_spec_cpu,dev_specs[i]);
			}else if(dev_specs[i].getDeviceType()==hype::GPU){
                            #ifdef ENABLE_GPU_ACCELERATION
				hype::Scheduler::instance().addAlgorithm(group_by_alg_spec_gpu,dev_specs[i]);		
#endif
			}
		}
		
		
		//hype::Scheduler::instance().addAlgorithm("SELECTION","GPU_Groupby_Algorithm",hype::GPU,"Least Squares 1D","Periodic Recomputation");
		map["CPU_Groupby_Algorithm"]=create_CPU_Groupby_Operator;
		map["GPU_Groupby_Algorithm"]=create_GPU_Groupby_Operator;
		return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
	}

	}//end namespace physical_operator

}//end namespace query_processing

}; //end namespace CogaDB
