
#include <query_processing/scan_operator.hpp>
#include <util/hardware_detector.hpp>

namespace CoGaDB
{

	namespace query_processing
	{
		//Map_Init_Function init_function_scan_operator=physical_operator::map_init_function_scan_operator;

		namespace physical_operator
		{

			TypedOperatorPtr create_scan_operator(TypedLogicalNode& logical_node, const hype::SchedulingDecision& sched_dec, TypedOperatorPtr, TypedOperatorPtr)
			{
				logical_operator::Logical_Scan& log_sort_ref = static_cast<logical_operator::Logical_Scan&>(logical_node);
				//std::cout << "create SCAN Operator!" << std::endl;
				/*
				if(!left_child) {
					std::cout << "Error!" << std::endl;
					exit(-1);
				}
				assert(right_child==NULL); //unary operator
				*/
				return TypedOperatorPtr( new scan_operator(sched_dec, log_sort_ref.getTablePtr()) );
			}



			Physical_Operator_Map_Ptr map_init_function_scan_operator()
			{
				//std::cout << sd.getNameofChoosenAlgorithm() << std::endl;
				Physical_Operator_Map map;
				if(!quiet) std::cout << "calling map init function! (SCAN OPERATION)" << std::endl;
				//hype::Scheduler::instance().addAlgorithm("SCAN","TABLE_SCAN",hype::CPU,"Least Squares 1D","Periodic Recomputation");
				hype::AlgorithmSpecification selection_alg_spec_cpu("TABLE_SCAN",
																	"SCAN",
																	hype::Least_Squares_1D,
																	hype::Periodic);


		//addAlgorithmSpecificationToHardware();

		const DeviceSpecifications& dev_specs = HardwareDetector::instance().getDeviceSpecifications();

		for(unsigned int i=0;i<dev_specs.size();++i){
			if(dev_specs[i].getDeviceType()==hype::CPU){
				hype::Scheduler::instance().addAlgorithm(selection_alg_spec_cpu,dev_specs[i]);
			}
		}

				map["TABLE_SCAN"]=create_scan_operator;
				//map["GPU_Algorithm"]=create_GPU_SORT_Operator;
				return Physical_Operator_Map_Ptr(new Physical_Operator_Map(map));
			}

			scan_operator::scan_operator(const hype::SchedulingDecision& sched_dec, TablePtr table)
                        : UnaryOperator<TablePtr,TablePtr>(sched_dec, TypedOperatorPtr()),table_(table)
				
			{
					this->result_=table_;
			}
			bool scan_operator::execute()
			{	if(!quiet && verbose && debug) std::cout << "Execute Scan"  << std::endl;
				//const TablePtr sort(TablePtr table, const std::string& column_name, SortOrder order=ASCENDING, MaterializationStatus mat_stat=MATERIALIZE, ComputeDevice comp_dev=CPU);
				this->result_=table_;
                if (this->result_) {
                    setResultSize(this->result_->getNumberofRows());
                    return true;
                } else
                    return false;
			}

			scan_operator::~scan_operator() {}

		}//end namespace physical_operator


		namespace logical_operator
		{
			Logical_Scan::Logical_Scan(std::string table_name) : TypedNode_Impl<TablePtr,physical_operator::map_init_function_scan_operator>(), table_(getTablebyName(table_name))
			{
                            if(table_==NULL){
                                COGADB_FATAL_ERROR(std::string("Table ") + table_name + " not found in global Table List!" ,"");
                            }
			}
                        Logical_Scan::Logical_Scan(TablePtr table) : TypedNode_Impl<TablePtr,physical_operator::map_init_function_scan_operator>(), table_(table)
			{
                            assert(table_!=NULL);
			}
			unsigned int Logical_Scan::getOutputResultSize() const
			{
				//TablePtr table_ptr = getTablebyName(table_name_);
				//assert(table_ptr != NULL);

				return table_->getNumberofRows();
			}

            double Logical_Scan::getCalculatedSelectivity() const {
				return 1.0;
			}
			std::string Logical_Scan::getOperationName() const
			{
				return "SCAN";
			}
                        
                        std::string Logical_Scan::toString(bool verbose) const{
				return std::string("SCAN ")+table_->getName();
			}
                        
			const std::string& Logical_Scan::getTableName()
			{
				return table_->getName();
			}
                        
                        const TablePtr Logical_Scan::getTablePtr()
			{
				return table_;
			}

	//redefining virtual function -> workaround so that scan operator can be processed uniformely with other operators
	TypedOperatorPtr Logical_Scan::getOptimalOperator(TypedOperatorPtr left_child, TypedOperatorPtr right_child, hype::DeviceTypeConstraint dev_constr){
		hype::Tuple t;

		t.push_back(getOutputResultSize());

		return this->operator_mapper_.getPhysicalOperator(*this, t, left_child, right_child, dev_constr); //this->getOperationName(), t, left_child, right_child);
	}

		}//end namespace logical_operator

	}//end namespace query_processing

}; //end namespace CogaDB
