#include <string>
#include <iostream>
#include <exception>

#include "sql/server/sql_parsetree.hpp"
#include "sql/server/sql_driver.hpp"
#include "sql_parser.hpp"

#define NDEBUG

namespace CoGaDB {
namespace SQL {

Driver::Driver()
	: buffer(NULL), istream(NULL), parser(*this, scanner)
{
	init_scan();

#ifndef NDEBUG
	parser.set_debug_level(1);
#endif
}

ParseTree::SequencePtr
Driver::parse(std::istream &is)
{
	set_input(is);
	if (parser.parse())
		throw ParseError();

	return result;
}

ParseTree::SequencePtr
Driver::parse(const std::string &src)
{
	set_input(src);
	if (parser.parse())
		throw ParseError();

	return result;
}

Driver::~Driver()
{
	destroy_scan();
}

bool
commandlineExec(const std::string &input)
{       Timestamp begin = getTimestamp();
	Driver driver;
	ParseTree::SequencePtr seq;
	TablePtr result;

	try {
		seq = driver.parse(input);
	} catch (const std::exception &e) {
		std::cerr << e.what() << std::endl;
		return false;
	}

	result = seq->execute();
        Timestamp end = getTimestamp();
        assert(end>=begin);
	if (result)
		result->print();
        //std::streamsize precision = std::cout.precision();
        //std::cout.precision(3);
        //std::cout << "Execution Time: "; // << double(end-begin)/(1000*1000) << "ms" << std::endl;
        //std::cout.precision(precision);
        printf("Execution Time: %.5lf ms\n", double(end-begin)/(1000*1000));
        //std::cout << std::endl;
	return true;
}

} /* namespace SQL */
} /* namespace CoGaDB */
